import pytest
import re
import json
import responses

from pytos2.securetrack.generic_ignored_interface import GenericIgnoredInterface
from pytos2.utils import get_api_node


class TestGenericIgnoredInterface:
    ignored_interfaces = json.load(
        open("tests/securetrack/json/generic_ignored_interfaces/mgmt-10.json")
    )
    interface = ignored_interfaces["IgnoredInterfaces"]

    @responses.activate
    def test_add_generic_ignored_interfaces_200(self, st):
        """Add one or multiple interface"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/ignoredinterface",
            status=201,
        )

        """Add single interface"""
        newInterface = st.add_generic_ignored_interface(
            interface_name=self.interface[0]["interfaceName"],
            device=self.interface[0]["mgmtId"],
            ip=self.interface[0]["ip"],
        )
        assert newInterface is None

    @responses.activate
    def test_add_generic_ignored_interfaces_400(self, st):
        """Add one or multiple interface"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/ignoredinterface",
            status=400,
        )

        """Add single interface"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_ignored_interface(
                interface_name=self.interface[0]["interfaceName"],
                device=self.interface[0]["mgmtId"],
                ip=self.interface[0]["ip"],
            )
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_ignored_interfaces_404(self, st):
        """Add one or multiple interface"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/ignoredinterface",
            status=404,
        )

        """Add single interface"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_ignored_interface(
                interface_name=self.interface[0]["interfaceName"],
                device=self.interface[0]["mgmtId"],
                ip=self.interface[0]["ip"],
            )
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_ignored_interfaces(self, st, generic_ignored_interface_mock):
        """Get interface by interface id"""
        interfaces = [
            GenericIgnoredInterface.kwargify(d)
            for d in get_api_node(
                self.ignored_interfaces, "IgnoredInterfaces", listify=True
            )
        ]
        interfaceByInt = st.get_generic_ignored_interfaces(10)
        assert interfaceByInt == interfaces

        with pytest.raises(ValueError) as exception:
            st.get_generic_ignored_interfaces(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_ignored_interfaces(
        self, st, generic_ignored_interface_mock
    ):
        """Delete interface by interface id"""
        interfaceByInt = st.delete_generic_ignored_interfaces(10)
        assert interfaceByInt is None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_ignored_interfaces(404)
        assert "Not Found" in str(exception.value)



from typing import Optional, TypedDict, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pytos2.secureapp.entrypoint import Sa
    from pytos2.securechange.entrypoint import Scw
    from pytos2.securetrack.entrypoint import St

Apps = Union["Sa", "Scw", "St"]


class APISessionKwargs(TypedDict):
    app: Apps
    hostname: Optional[str]
    username: Optional[str]
    password: Optional[str]
    scheme: Optional[str]
    verify: Optional[bool]


class OAuthToken(TypedDict):
    access_token: str
    refresh_token: str
    token_type: str
    expires_in: str
    expires_at: str
    mac_key: str
    mac_algorithm: str



import pytest
import responses

from pytos2.securechange.workflow_triggers import (
    Script,
    Workflow,
    Trigger,
    WorkflowTrigger,
)


class TestWorkflowTrigger:
    @responses.activate
    def test_get_triggers(self, workflow_triggers_mock, scw):
        workflow_triggers = scw.get_triggers()

        assert workflow_triggers[0].name == "Test Integration"
        assert workflow_triggers[0].script.path == ""
        assert (
            workflow_triggers[0].script.arguments
            == "/opt/tufin/securitysuite/ps/bin/attach_test_report_to_ticket.py"
        )
        assert workflow_triggers[0].triggers[0].name == "New trigger group"
        assert workflow_triggers[0].triggers[0].workflow.name == "AR_MG"
        assert workflow_triggers[0].triggers[0].workflow.parent_workflow_id == 426
        assert workflow_triggers[0].triggers[0].events == ["CREATE", "ADVANCE"]

        assert workflow_triggers[1].name == "Test One"
        assert (
            workflow_triggers[1].script.path
            == "test-one-solutions/populate-access-request"
        )
        assert workflow_triggers[1].script.arguments == ""
        assert workflow_triggers[1].triggers[0].name == "Test One"
        assert workflow_triggers[1].triggers[0].workflow.name == "Test One"
        assert workflow_triggers[1].triggers[0].workflow.parent_workflow_id == 595
        assert workflow_triggers[1].triggers[0].events == ["REOPEN"]

    @responses.activate
    def test_get_trigger(self, workflow_triggers_mock, scw):
        workflow_trigger = scw.get_trigger(id=4)

        assert workflow_trigger.name == "Test Integration"
        assert workflow_trigger.script.path == ""
        assert (
            workflow_trigger.script.arguments
            == "/opt/tufin/securitysuite/ps/bin/attach_test_report_to_ticket.py"
        )
        assert workflow_trigger.triggers[0].name == "New trigger group"
        assert workflow_trigger.triggers[0].workflow.name == "AR_MG"
        assert workflow_trigger.triggers[0].workflow.parent_workflow_id == 426
        assert workflow_trigger.triggers[0].events == ["CREATE", "ADVANCE"]

        workflow_trigger = scw.get_trigger(name="Contoso")

        assert workflow_trigger.name == "Test One"
        assert (
            workflow_trigger.script.path == "test-one-solutions/populate-access-request"
        )
        assert workflow_trigger.script.arguments == ""
        assert workflow_trigger.triggers[0].name == "Test One"
        assert workflow_trigger.triggers[0].workflow.name == "Test One"
        assert workflow_trigger.triggers[0].workflow.parent_workflow_id == 595
        assert workflow_trigger.triggers[0].events == ["REOPEN"]

        with pytest.raises(ValueError) as e:
            scw.get_trigger()

        with pytest.raises(ValueError) as e:
            scw.get_trigger(id=4, name="Test One")

    @responses.activate
    def test_add_trigger(self, workflow_triggers_mock, scw):
        script = Script(path="test-solution/test-solution", arguments="")
        workflow = Workflow(name="test workflow", parent_workflow_id=1)
        trigger = Trigger(
            name="test trigger", workflow=workflow, events=["CREATE", "ADVANCE"]
        )
        workflow_trigger = WorkflowTrigger(
            name="test", script=script, triggers=[trigger]
        )
        scw.add_trigger(workflow_trigger)



import json

import pytest
import responses
from requests.exceptions import HTTPError
from . import conftest  # noqa
from pytos2.securechange.ticket import Ticket, TicketStatus, Step, Task
from pytos2.securechange.fields import (
    TextArea,
    TextField,
    Checkbox,
    FieldXsiType,
    MultiAccessRequest,
)


class TestTicketAttributes(object):
    def test_steps(self, all_fields_ticket):
        assert isinstance(all_fields_ticket.steps, list)

    def test_workflow(self, all_fields_ticket):
        workflow = all_fields_ticket.workflow

        assert workflow.name == "AR all fields"

    def test_current_step_props(self, all_fields_ticket):
        assert all_fields_ticket.current_step_index == 3
        assert all_fields_ticket.current_step_name == "No Fields"

        assert all_fields_ticket.previous_step.name == "Duplicate"

        empty_ticket = Ticket.create("test_workflow", "test_step")
        assert empty_ticket.current_step_index is None
        assert empty_ticket.previous_step is None

    @pytest.mark.parametrize(
        "attr, value",
        [
            ("id", 288),
            ("subject", "Better Tests"),
            ("requester", "r"),
            ("requester_id", 4),
            ("priority", Ticket.Priority.NORMAL),
            ("status", TicketStatus.INPROGRESS),
            ("expiration_field_name", "Expiration"),
            ("expiration_date", None),
            ("domain_name", "Default"),
            ("sla_status", Ticket.SlaStatus.NA),
            ("sla_outcome", Ticket.SlaOutcome.NA),
        ],
    )
    def test_values(self, all_fields_ticket, attr, value):
        assert getattr(all_fields_ticket, attr) == value

    def test_last_step(self, all_fields_ticket, scw):
        assert all_fields_ticket.last_step.name == "No Fields"

    def test_current_step(self, all_fields_ticket, closed_group_modify_ticket):
        assert (
            all_fields_ticket.current_step_name
            == all_fields_ticket.last_step.name
            == all_fields_ticket.data["ticket"]["current_step"]["name"]
        )
        assert closed_group_modify_ticket.current_step_name is None

    def test_access_request(
        self, open_access_request_ticket, closed_group_modify_ticket
    ):
        assert open_access_request_ticket.access_request
        assert closed_group_modify_ticket.access_request is None

    def test_group_modify(self, open_group_modify_ticket, closed_group_modify_ticket):
        assert open_group_modify_ticket.group_modify
        assert closed_group_modify_ticket.group_modify is None

    def test_last_task(self, all_fields_ticket):
        assert all_fields_ticket.last_task is all_fields_ticket.steps[-1].tasks[-1]

    def test_current_task(self, all_fields_ticket):
        assert all_fields_ticket.current_task is all_fields_ticket.steps[-1].tasks[-1]

    def test_current_task_dynamic_assignment(self, dynamic_assignment_ticket):
        with pytest.raises(ValueError):
            dynamic_assignment_ticket.current_task

    def test_current_tasks(self, all_fields_ticket):
        assert all_fields_ticket.current_tasks is all_fields_ticket.steps[-1].tasks

    def test_json(self, all_fields_ticket):
        assert isinstance(json.dumps(all_fields_ticket._json), str)

    def test_application_details(self, application_details):
        assert application_details.id == 8


class TestTicketFunctions(object):
    def test_create_ticket(self):
        ticket = Ticket.create("workflow", "test subject")
        assert ticket.subject == "test subject"
        assert ticket.workflow.name == "workflow"

        ticket.create_step("Step Name")
        assert ticket.steps[0].name == "Step Name"
        assert isinstance(ticket.steps[0], Step)

    def test_get_step(self, all_fields_ticket):
        assert (
            all_fields_ticket.get_step("invalid")
            is all_fields_ticket.get_step(100)
            is None
        )
        assert (
            all_fields_ticket.get_step("Open request")
            == all_fields_ticket.get_step(0)
            == all_fields_ticket.steps[0]
        )

    def test_get_step_by_id(self, all_fields_ticket):
        assert all_fields_ticket._get_step(1590).id == 1590

    def test_step_task(self, all_fields_ticket):
        assert (
            all_fields_ticket.get_step_task("invalid", 0)
            is all_fields_ticket.get_step_task("invalid", "invalid")
            is None
        )
        assert all_fields_ticket.get_step_task("invalid", 0) is None

    def test_step_task_fields(self, all_fields_ticket):
        assert all_fields_ticket.get_step_task_fields("invalid", 0) is None
        assert not all_fields_ticket.get_step_task_fields(1)

    @responses.activate
    def test_advance(self, first_step_ticket, redo_ticket, closed_ticket):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/252/steps/1246/tasks/1258",
        )
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/252",
            json=closed_ticket.data,
        )
        res = redo_ticket.advance()
        assert isinstance(res, Ticket)
        first_step_ticket.advance(save=False)
        assert all(t.is_done for t in first_step_ticket.current_step.tasks)

    def test_advance_closed(self, closed_group_modify_ticket):
        with pytest.raises(AssertionError):
            closed_group_modify_ticket.advance()

    def test_save_new(self, all_fields_ticket):
        all_fields_ticket.id = None
        with pytest.raises(AssertionError):
            all_fields_ticket.save()

    @responses.activate
    def test_save_exceptions(self, all_fields_ticket, scw):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/steps/current/tasks/1625",
            status=406,
        )
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288/steps/current/tasks/1625/fields",
            status=406,
        )

        new_field = all_fields_ticket.last_task.create_field(
            "TestField", FieldXsiType.TEXT_FIELD
        )
        new_field.text = "123"
        with pytest.raises(ValueError):
            all_fields_ticket.save()

        with pytest.raises(ValueError):
            all_fields_ticket.save(force=True)

        all_fields_ticket.last_task.done()
        with pytest.raises(ValueError):
            all_fields_ticket.save()

    @responses.activate
    def test_save_unchanged(self, all_fields_ticket):
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288",
            json={"ticket": {**all_fields_ticket.data["ticket"], "id": 288}},
        )
        assert isinstance(all_fields_ticket.save(), Ticket)

    @responses.activate
    def test_save_field(self, open_access_request_ticket):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/242/steps/current/tasks/1216/fields",
        )
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/242",
            json={"ticket": {**open_access_request_ticket.data["ticket"], "id": 242}},
        )

        open_access_request_ticket.last_task.get_field("Design comment").text = (
            "new text"
        )
        assert isinstance(open_access_request_ticket.save(), Ticket)

    @responses.activate
    def test_save_task(self, open_access_request_ticket):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/242/steps/current/tasks/1216",
        )
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/242",
            json={"ticket": {**open_access_request_ticket.data["ticket"], "id": 242}},
        )
        open_access_request_ticket.last_task.status = Task.Status.DONE
        open_access_request_ticket.current_step.tasks.append(Task())

        assert isinstance(open_access_request_ticket.save(), Ticket)

    @responses.activate
    def test_save_task_force(self, open_access_request_ticket):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/242/steps/current/tasks/1216",
        )
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/242",
            json={"ticket": {**open_access_request_ticket.data["ticket"], "id": 242}},
        )
        open_access_request_ticket.last_task.status = Task.Status.DONE

        assert isinstance(open_access_request_ticket.save(force=True), Ticket)

    @responses.activate
    def test_redo(self, closed_ticket, first_step_ticket, redo_ticket):
        counter = {"count": 0}

        def ticket_cb(c):
            def _f(request):
                c["count"] += 1
                if c["count"] < 2:
                    return (200, {}, json.dumps(redo_ticket.data))
                else:
                    return (200, {}, json.dumps(first_step_ticket.data))

            return _f

        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/252/steps/1246/tasks/1258/redo/1245",
        )
        responses.add_callback(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/252",
            callback=ticket_cb(counter),
        )
        res = redo_ticket.redo(0)
        assert len(res.steps) == 1

    @responses.activate
    def test_redo_failed(self, closed_ticket, first_step_ticket, redo_ticket):
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/252/steps/1246/tasks/1258/redo/1245",
            status=400,
        )
        with pytest.raises(HTTPError):
            redo_ticket.redo(0)

    def test_redo_closed(self, closed_ticket):
        with pytest.raises(AssertionError):
            closed_ticket.redo(0)

    def test_redo_missing_step(self, redo_ticket):
        with pytest.raises(IndexError):
            redo_ticket.redo(10000)

    def test_redo_first_step(self, first_step_ticket):
        with pytest.raises(AssertionError):
            first_step_ticket.redo(0)

    def test_redo_no_step_id(self, redo_ticket):
        redo_ticket.steps[0].id = None
        with pytest.raises(AssertionError):
            redo_ticket.redo(0)

    def test_redo_current_step(self, redo_ticket):
        with pytest.raises(AssertionError):
            redo_ticket.redo(1)

    def test_misc_redo_errors(self, redo_ticket):
        redo_ticket.current_task.id = None
        with pytest.raises(AssertionError):
            redo_ticket.redo(0)
        redo_ticket.current_step.id = None
        with pytest.raises(AssertionError):
            redo_ticket.redo(0)
        redo_ticket.id = None
        with pytest.raises(AssertionError):
            redo_ticket.redo(0)

    def test_save_closed(self, closed_group_modify_ticket, scw):
        with pytest.raises(AssertionError):
            closed_group_modify_ticket.save()

    def test_json_override(self, all_fields_ticket):
        all_fields_ticket._json = {}
        assert all_fields_ticket._json == {}

    def test_post_json(self, all_fields_ticket):
        new_json = all_fields_ticket._json
        new_json["ticket"]["steps"]["step"] = [new_json["ticket"]["steps"]["step"][0]]
        assert all_fields_ticket.post_json == new_json

    @responses.activate
    def test_post(self, all_fields_ticket, scw):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets",
            headers={
                "Location": "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/614"
            },
        )
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/614",
            json={"ticket": {**all_fields_ticket.data["ticket"], "id": 614}},
        )

        t = all_fields_ticket.post()
        assert t.id == 614

    @responses.activate
    def test_bad_post(self, all_fields_ticket, scw, post_bad_ticket_json):
        responses.add(
            responses.POST,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets",
            status=400,
            json=post_bad_ticket_json,
        )
        with pytest.raises(HTTPError):
            all_fields_ticket.post()

    @responses.activate
    def test_reject(self, all_fields_ticket, scw):
        responses.add(
            responses.PUT,
            f"https://198.18.0.1/securechangeworkflow/api/securechange/tickets/{all_fields_ticket.id}/reject",
        )
        responses.add(
            responses.GET,
            f"https://198.18.0.1/securechangeworkflow/api/securechange/tickets/{all_fields_ticket.id}",
            json=all_fields_ticket.data,
        )
        assert isinstance(all_fields_ticket.reject(), Ticket)
        all_fields_ticket.id = None
        with pytest.raises(ValueError):
            all_fields_ticket.reject()

    @responses.activate
    def test_reject_not_supported(self, all_fields_ticket, scw):
        responses.add(
            responses.PUT,
            f"https://198.18.0.1/securechangeworkflow/api/securechange/tickets/{all_fields_ticket.id}/cancel",
        )
        responses.add(
            responses.PUT,
            f"https://198.18.0.1/securechangeworkflow/api/securechange/tickets/{all_fields_ticket.id}/reject",
            json={
                "result": {
                    "code": "WEB_APPLICATION_ERROR",
                    "message": "HTTP 404 Not Found",
                }
            },
            status=400,
        )
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/1/reject",
            json={"result": {"code": "WEB_APPLICATION_ERROR"}},
            status=400,
        )
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/2/reject",
            json={"result": {"code": "WEB_APPLICATION_ERROR", "message": "scarry"}},
            status=400,
        )
        responses.add(
            responses.GET,
            f"https://198.18.0.1/securechangeworkflow/api/securechange/tickets/{all_fields_ticket.id}",
            json=all_fields_ticket.data,
        )
        with pytest.raises(HTTPError):
            all_fields_ticket.reject()
        all_fields_ticket.id = 1
        with pytest.raises(HTTPError):
            all_fields_ticket.reject()
        all_fields_ticket.id = 2
        with pytest.raises(HTTPError):
            all_fields_ticket.reject()


class TestStep(object):
    @pytest.fixture()
    def current_step(self, all_fields_ticket):
        return all_fields_ticket.current_step

    @pytest.fixture()
    def first_step(self, all_fields_ticket):
        return all_fields_ticket.get_step(0)

    @pytest.fixture()
    def all_fields_step(self, all_fields_ticket):
        return all_fields_ticket.get_step(1)

    def test_step_index(self, all_fields_ticket, all_fields_step_json):
        assert (
            all_fields_ticket.get_step(3).name
            == all_fields_ticket.get_step(-1).name
            == all_fields_step_json[-1]
        )

    def test_create_task(self):
        step = Step()

        assert len(step.tasks) == 0
        step.create_task()
        assert len(step.tasks) == 1
        assert isinstance(step.tasks[0], Task)

    def test_task(self, all_fields_step):
        assert all_fields_step.get_task(0).name is None
        assert all_fields_step.get_task(None).name is None
        assert all_fields_step.get_task("missing") is None
        assert all_fields_step.get_task(100) is None

    def test_task_field(self, all_fields_step):
        fields = all_fields_step.get_task_fields(0, TextArea)
        print(fields)

        assert len(fields) == 2
        assert isinstance(fields[0], TextArea)

        fields = all_fields_step.get_task_fields(0, MultiAccessRequest)
        assert len(fields) == 1
        assert isinstance(fields[0], MultiAccessRequest)

        fields = all_fields_step.get_task_fields(0)
        assert len(fields) == 0

    def test_done(self, current_step):
        stat = not all(t.status is Task.Status.DONE for t in current_step.tasks)
        assert stat  # all(t.status is Task.Status.DONE for t in current_step.tasks) is False
        current_step.done()
        assert all(t.status is Task.Status.DONE for t in current_step.tasks)

    def test_is_done(self, first_step):
        assert first_step.is_done


class TestTask(object):
    @pytest.fixture()
    def task(self, all_fields_ticket):
        return all_fields_ticket.get_step_task(1)

    @pytest.fixture()
    def open_task(self, all_fields_ticket):
        return all_fields_ticket.last_task

    def test_create_field(self):
        task = Task()

        assert len(task.fields) == 0
        task.create_field("TestField", TextArea)

        assert len(task.fields) == 1
        assert isinstance(task.fields[0], TextArea)
        assert task.fields[0].name == "TestField"

        task.fields = []
        task.create_field("TestField2", FieldXsiType.TEXT_FIELD)

        assert len(task.fields) == 1
        assert isinstance(task.fields[0], TextField)
        assert task.fields[0].name == "TestField2"

    def test_fields(self, task):
        assert all(
            t.xsi_type
            in (FieldXsiType.TEXT_FIELD, FieldXsiType.TEXT_AREA, FieldXsiType.CHECKBOX)
            for t in task._get_fields(TextField, TextArea, Checkbox)
        )

    def test_get_field(self, task):
        assert task.get_field("Text area")
        assert task.get_field("Text area", TextArea)
        assert isinstance(task.get_field(None, TextArea), TextArea)
        assert task.get_field("Text area", TextField) is None
        assert task.get_field("Text field", TextField, TextArea)

    def test_dirty(self, open_task):
        open_task.done()
        assert open_task._dirty

    def test_dirty_fields(self, task):
        f = task.get_field("Approve / Reject")
        f.approve()
        assert task._dirty_fields


class TestTicketIterator:
    @responses.activate
    def test_iterator(self, scw, tickets_mock):
        # looks like ticket_mock can't differenciate different params, so I changed it to a dummy url
        tickets = scw.get_tickets(TicketStatus.CLOSED)

        assert len(tickets) == 200
        assert isinstance(tickets, list)
        assert isinstance(tickets[0], Ticket)


class TestTicketLifecycle:
    @responses.activate
    def test_change_requester(self, ticket_life_cycle_mock, scw):
        ticket = scw.get_ticket(288)
        user = scw.get_user(45)

        scw.change_requester(ticket, user, "change requester")
        scw.change_requester(288, 45, "change requester")

        ticket.change_requester(user, "change requester")
        ticket.change_requester(45, "change requester")

    @responses.activate
    def test_reassign_ticket(self, ticket_life_cycle_mock, scw):
        ticket = scw.get_ticket(288)
        step = ticket.current_step
        task = step.last_task
        user = scw.get_user(45)

        scw.reassign_ticket(ticket, user, step, task, "reassign ticket")
        scw.reassign_ticket(288, 45, step, task, "reassign ticket")

        ticket.reassign(user, step, task, "reassign ticket")
        ticket.reassign(45, step, task, "reassign ticket")

    @responses.activate
    def test_reject_ticket(self, ticket_life_cycle_mock, scw):
        ticket = scw.get_ticket(288)
        ticket.reject()
        ticket.reject(comment="Customer comment")
        ticket.reject(comment="Customer comment", handler_id=45)

    @responses.activate
    def test_cancel_ticket(self, ticket_life_cycle_mock, scw):
        ticket = scw.get_ticket(288)
        ticket.cancel()
        ticket.cancel(requestor_id=45)

    @responses.activate
    def test_confirm_ticket(self, ticket_life_cycle_mock, scw):
        ticket = scw.get_ticket(288)
        ticket.confirm()
        ticket.confirm(comment="Customer comment")
        ticket.confirm(comment="Customer comment", requestor_id=45)

    @responses.activate
    def test_get_ticket_events(self, ticket_life_cycle_ticket_events_mock, scw):
        ticket_events = scw.get_ticket_events()

        assert ticket_events[0].ticket_id == 1036
        assert ticket_events[0].workflow_name == "Firewall Change Request Workflow"
        assert ticket_events[0].domain_id == 1
        assert ticket_events[0].parent_workflow_id == 324
        assert ticket_events[0].timestamp.strftime("%Y-%m-%d") == "2024-04-25"
        assert ticket_events[0].type == "TICKET_SLA_OVERDUE"
        assert ticket_events[0].step_id is None
        assert ticket_events[0].step_name is None
        assert ticket_events[0].task_id is None
        assert ticket_events[0].task_name is None
        assert len(ticket_events[0].potential_participants) == 0
        assert ticket_events[0].task_assigned_data is None

        assert ticket_events[1].ticket_id == 1036
        assert ticket_events[1].workflow_name == "Firewall Change Request Workflow"
        assert ticket_events[1].domain_id == 1
        assert ticket_events[1].parent_workflow_id == 324
        assert ticket_events[1].timestamp.strftime("%Y-%m-%d") == "2024-02-18"
        assert ticket_events[1].type == "TASK_READY_TO_BE_ASSIGNED"
        assert ticket_events[1].step_id == 5679
        assert ticket_events[1].step_name == "Business Approval"
        assert ticket_events[1].task_id == 5695
        assert ticket_events[1].task_name == "Default"
        assert ticket_events[1].potential_participants[0].participant_id == 9
        assert (
            ticket_events[1].potential_participants[0].participant_name
            == "Security Team"
        )
        assert ticket_events[1].potential_participants[1].participant_id == 4
        assert ticket_events[1].potential_participants[1].participant_name == "r"
        assert ticket_events[1].task_assigned_data is None

        assert ticket_events[2].ticket_id == 1036
        assert ticket_events[2].workflow_name == "Firewall Change Request Workflow"
        assert ticket_events[2].domain_id == 1
        assert ticket_events[2].parent_workflow_id == 324
        assert ticket_events[2].timestamp.strftime("%Y-%m-%d") == "2024-02-18"
        assert ticket_events[2].type == "TASK_ASSIGNED"
        assert ticket_events[2].step_id == 5679
        assert ticket_events[2].step_name == "Business Approval"
        assert ticket_events[2].task_id == 5695
        assert ticket_events[2].task_name == "Default"
        assert len(ticket_events[2].potential_participants) == 0
        assert ticket_events[2].task_assigned_data.assignee_id == 4
        assert ticket_events[2].task_assigned_data.assignee_name == "r"
        assert ticket_events[2].task_assigned_data.participant_id == 9
        assert ticket_events[2].task_assigned_data.participant_name == "Security Team"

        assert ticket_events[3].ticket_id == 1031
        assert ticket_events[3].workflow_name == "Access Request Workflow"
        assert ticket_events[3].domain_id == 1
        assert ticket_events[3].parent_workflow_id == 322
        assert ticket_events[3].timestamp.strftime("%Y-%m-%d") == "2024-02-15"
        assert ticket_events[3].type == "TASK_DONE"
        assert ticket_events[3].step_id == 5645
        assert ticket_events[3].step_name == "Identify Targets and Risks"
        assert ticket_events[3].task_id == 5661
        assert ticket_events[3].task_name == "Default"
        assert len(ticket_events[3].potential_participants) == 0
        assert ticket_events[3].task_assigned_data is None

    @responses.activate
    def test_get_ticket_events_by_filter(
        self, ticket_life_cycle_ticket_events_mock, scw
    ):
        ticket_events = scw.get_ticket_events(type="TICKET_SLA_OVERDUE")
        assert len(ticket_events) == 1
        assert ticket_events[0].type == "TICKET_SLA_OVERDUE"

        ticket_events = scw.get_ticket_events(assignee_name="r")
        assert len(ticket_events) == 1
        assert ticket_events[0].task_assigned_data.assignee_name == "r"

        ticket_events = scw.get_ticket_events(assignee_id=4)
        assert len(ticket_events) == 1
        assert ticket_events[0].task_assigned_data.assignee_id == 4

        ticket_events = scw.get_ticket_events(participant_name="Security Team")
        assert len(ticket_events) == 1
        assert ticket_events[0].task_assigned_data.participant_name == "Security Team"

        ticket_events = scw.get_ticket_events(participant_id=9)
        assert len(ticket_events) == 1
        assert ticket_events[0].task_assigned_data.participant_id == 9

        ticket_events = scw.get_ticket_events(ticket_id=1031)
        assert len(ticket_events) == 1
        assert ticket_events[0].ticket_id == 1031

        ticket_events = scw.get_ticket_events(workflow_name="Access Request Workflow")
        assert len(ticket_events) == 1
        assert ticket_events[0].workflow_name == "Access Request Workflow"

        ticket_events = scw.get_ticket_events(parent_workflow_id=322)
        assert len(ticket_events) == 1
        assert ticket_events[0].parent_workflow_id == 322

        ticket_events = scw.get_ticket_events(step_name="Identify Targets and Risks")
        assert len(ticket_events) == 1
        assert ticket_events[0].step_name == "Identify Targets and Risks"

        ticket_events = scw.get_ticket_events(step_id=5645)
        assert len(ticket_events) == 1
        assert ticket_events[0].step_id == 5645

        ticket_events = scw.get_ticket_events(date_from="2024-04-25")
        assert len(ticket_events) == 1
        assert ticket_events[0].timestamp.strftime("%Y-%m-%d") == "2024-04-25"

        ticket_events = scw.get_ticket_events(date_to="2024-02-16")
        assert len(ticket_events) == 1
        assert ticket_events[0].timestamp.strftime("%Y-%m-%d") == "2024-02-15"

    @responses.activate
    def test_backfill_ticket_events(self, ticket_life_cycle_ticket_events_mock, scw):
        scw.backfill_ticket_events("2023-11-05")

    @responses.activate
    def test_get_ticket_historical_events_status(
        self, ticket_life_cycle_ticket_events_mock, scw
    ):
        historical_events_status = scw.get_ticket_historical_events_status()
        assert historical_events_status == "NOT_RUNNING"

    @responses.activate
    def test_map_rules(self, ticket_life_cycle_mock, scw):
        scw.map_rules(1037)
        scw.map_rules(1037, 5)

        ticket = scw.get_ticket(1037)
        ticket.map_rules()

    @responses.activate
    def test_designer_redesign(self, ticket_life_cycle_mock, scw):
        scw.designer_redesign(1037, 5689, 5705, 61671)

        ticket = scw.get_ticket(1037)
        ticket.designer_redesign()

    @responses.activate
    def test_designer_device_commit(self, ticket_life_cycle_mock, scw):
        scw.designer_device_commit(1037, 5689, 5705, 61671, 8)

        ticket = scw.get_ticket(1037)
        ticket.designer_device_commit(8)

    @responses.activate
    def test_designer_device_update(self, ticket_life_cycle_mock, scw):
        scw.designer_device_update(1037, 5689, 5705, 61671, 8)
        scw.designer_device_update(1037, 5689, 5705, 61671, 8, True)

        ticket = scw.get_ticket(1037)
        ticket.designer_device_update(8)



from copy import deepcopy
from datetime import datetime
from enum import Enum
from typing import List, Optional, Union, Type, Iterator
from typing_extensions import Literal
from time import sleep

from requests import Response
from requests.exceptions import HTTPError
import attr

from traversify import Traverser

# avoid circular imports
import pytos2  # noqa
from pytos2.securechange.entrypoint import Scw
from pytos2.utils import (
    propify,
    kwargify,
    prop,
    jsonify,
    safe_date,
    safe_iso8601_date,
    get_api_node,
    TimeFormat,
    safe_unwrap_msg,
)

from .fields import (
    Link,
    classify_field,
    get_field_class,
    Field,
    MultiGroupChange,
    MultiAccessRequest,
    FieldXsiType,
    FieldType,
)

from .user import SCWUser

from pytos2.models import Jsonable, UnMapped


@propify
class Workflow(Jsonable):
    class Meta(Enum):
        ROOT = "workflow"

    class Prop(Enum):
        ID = "id"
        NAME = "name"
        USES_TOPOLOGY = "uses_topology"

    name: Optional[str] = prop(None)
    uses_topology: Optional[bool] = prop(None, jsonify=False)
    id: Optional[int] = prop(None, cmp=False, jsonify=False)


@propify
class ApplicationDetails(Jsonable):
    class Meta(Enum):
        ROOT = "application_details"

    class Prop(Enum):
        ID = "id"
        NAME = "name"
        DISPLAY_NAME = "display_name"
        LINK = "link"

    id: Optional[int] = prop(None, cmp=False, jsonify=False)
    name: Optional[str] = prop(None)
    display_name: Optional[str] = prop(None)
    link: Optional[Link] = prop(None, repr=False)


@propify
class Task(Jsonable):
    class Meta(Enum):
        ROOT = "task"

    class Prop(Enum):
        ID = "id"
        NAME = "name"
        STATUS = "status"
        ASSIGNEE = "assignee"
        ASSIGNEE_ID = "assignee_id"
        PENDING_REASON = "pending_reason"
        PENDING_REASON_DESCRIPTION = "pending_reason_description"
        FIELDS = "fields"

    class Status(Enum):
        WAITING_TO_BE_ASSIGNED = "WAITING_TO_BE_ASSIGNED"
        ASSIGNED = "ASSIGNED"
        WAITING_FOR_MORE_INFO = "WAITING_FOR_MORE_INFO"
        DONE = "DONE"
        INVALID = "INVALID"
        PENDING = "PENDING"
        PENDING_LICENSE = "PENDING_LICENSE"

    fields: List[FieldType] = prop(
        factory=list, repr=False, flatify=Field.Meta.ROOT.value, kwargify=classify_field
    )
    id: Optional[int] = prop(None, cmp=False)
    name: Optional[str] = prop(None, repr=False)
    status: Optional[Status] = prop(None)
    assignee: Optional[str] = prop(None, jsonify=False)
    assignee_id: Optional[str] = prop(None, repr=False, jsonify=False)
    pending_reason: Optional[str] = prop(None, repr=False, jsonify=False)
    pending_reason_description: Optional[str] = prop(None, repr=False, jsonify=False)

    def _get_fields(self, *types: Type[Field]) -> List[FieldType]:
        return [f for f in self.fields if isinstance(f, types)]

    def create_field(self, name: str, field_type: Union[Type[Field], FieldXsiType]):
        if isinstance(field_type, FieldXsiType):
            cls = get_field_class(field_type.value)
        else:
            cls = field_type

        field = cls.kwargify({"name": name})

        self.fields.append(field)
        return field

    def get_field(
        self, name: Optional[str] = None, *types: Type[Field]
    ) -> Optional[FieldType]:
        field = None
        for field in self.fields:
            field_name = getattr(field, "name", None) or field._json.get("name")
            if field_name == name:
                if types:
                    if isinstance(field, types):
                        return field
                else:
                    return field
            if name is None:
                if isinstance(field, types):
                    return field
        return None

    @property
    def group_modify(self) -> Optional[MultiGroupChange]:
        return self.get_field(None, MultiGroupChange)  # type: ignore

    @property
    def access_request(self) -> Optional[MultiAccessRequest]:
        return self.get_field(None, MultiAccessRequest)  # type: ignore

    def done(self) -> None:
        self.status = self.Status.DONE

    @property
    def is_done(self) -> bool:
        return self.status is self.Status.DONE

    @property
    def _dirty(self) -> bool:
        return (
            False
            if not self.status
            else True if not self.data else self.status.value != self.data.get("status")
        )

    @property
    def _dirty_fields(self) -> List:
        return [
            f
            for f in self.fields
            if isinstance(f, UnMapped) or (not f.read_only and f._dirty)
        ]

    @property
    def _json(self) -> dict:
        if self._json_override is not None:
            return self._json_override  # pragma: no cover
        j = jsonify(self)
        if not self.fields:
            j["fields"] = {}
        return j


@propify
class Step(Jsonable):
    class Meta(Enum):
        ROOT = "step"

    class Prop(Enum):
        ID = "id"
        NAME = "name"
        REDONE = "redone"
        SKIPPED = "skipped"
        TASKS = "tasks"

    id: Optional[int] = prop(None, cmp=False, jsonify=False)
    name: Optional[str] = prop(None)
    redone: Optional[bool] = prop(None, jsonify=False)
    skipped: Optional[bool] = prop(None, jsonify=False)
    tasks: List[Task] = prop(factory=list, repr=False, flatify=Task.Meta.ROOT.value)

    def get_task(self, identifier):
        task = None
        if isinstance(identifier, int):
            try:
                task = self.tasks[identifier]
            except IndexError:
                return None
        else:
            for task in self.tasks:
                if task.name == identifier:
                    break
            else:
                return None
        return task

    def get_task_fields(self, identifier, *types: Type[Field]):
        task = self.get_task(identifier)
        return task._get_fields(*types)

    def create_task(self):
        task = Task()
        self.tasks.append(task)
        return task

    @property
    def last_task(self) -> Task:
        return self.get_task(-1)

    def done(self):
        for task in self.tasks:
            task.done()

    @property
    def is_done(self):
        return all(task.is_done for task in self.tasks)


@propify
class CurrentStep(Jsonable):
    class Meta(Enum):
        ROOT = "current_step"

    class Prop(Enum):
        ID = "id"
        NAME = "name"

    name: Optional[str] = prop(None)
    id: Optional[int] = prop(None, cmp=False, jsonify=False)


class TicketStatus(Enum):
    ALL = "__ALL__"
    CLOSED = "Ticket Closed"
    CANCELLED = "Ticket Cancelled"
    REJECTED = "Ticket Rejected"
    RESOLVED = "Ticket Resolved"
    INPROGRESS = "In Progress"


@propify
class Attachment(Jsonable):
    uid: str
    name: str
    link: str = prop(None, flatify="@href")


@propify
class Comment(Jsonable):
    class Prop(Enum):
        USERS = "user"
        ATTACHMENT = "attachment"

    class CommentType(Enum):
        COMMENT = "comment"

    content: Optional[str] = prop(None)
    created: Optional[datetime] = prop(None, kwargify=safe_iso8601_date)
    task_name: Optional[str] = prop(None)
    commentType: Optional[CommentType] = prop(None)
    users: List[str] = prop(None, key=Prop.USERS.value)
    attachments: List[Attachment] = prop(
        None, jsonify=False, flatify=Prop.ATTACHMENT.value
    )


@propify
class TicketHistoryEntry(Jsonable):
    id: Optional[int] = prop(None, repr=False)

    date: Optional[datetime] = prop(
        None,
        kwargify=safe_iso8601_date,
        jsonify=lambda d: d.isoformat() if d else None,
        repr=lambda v: v.isoformat() if v else "None",
    )

    performed_by: Optional[str] = prop(None)
    description: Optional[str] = prop(None)
    step_name: Optional[str] = prop(None)
    task_name: Optional[str] = prop(None)


@propify
class PushCommandResult(Jsonable):
    class Status(Enum):
        SUCCEEDED = "SUCCEEDED"

    status: Optional[Status] = prop(None)
    order: int = prop()
    command: Optional[str] = prop(None)
    failure_reason: Optional[str] = prop(None)


@propify
class DesignerUpdateDeviceHistory(TicketHistoryEntry):
    vendor_name: Optional[str] = prop(None)
    device_name: Optional[str] = prop(None)

    push_command_results: List[PushCommandResult] = prop(
        flatify="push_command_result", factory=list, repr=False
    )


def get_ticket_history_entry_class(obj):
    if "@xsi.type" in obj:
        return {"designerUpdateDeviceHistoryDTO": DesignerUpdateDeviceHistory}.get(
            obj["@xsi.type"], TicketHistoryEntry
        )
    else:
        return TicketHistoryEntry


def classify_ticket_history_entry(obj):
    return get_ticket_history_entry_class(obj).kwargify(obj)


def _summarize_history_activity(v):
    if isinstance(v, list):
        return f"[... list with {len(v)} entries ...]"
    return repr(v)


@propify
class TicketHistory(Jsonable):
    ticket_id: int = prop()
    ticket_history_activity: List[TicketHistoryEntry] = prop(
        factory=list,
        kwargify=classify_ticket_history_entry,
        repr=_summarize_history_activity,
    )


@propify
class Ticket(Jsonable):
    """
    This class represents a SecureChange ticket object.
    """

    class Meta(Enum):
        ROOT = "ticket"

    class Priority(Enum):
        CRITICAL = "Critical"
        HIGH = "High"
        NORMAL = "Normal"
        LOW = "Low"

    class SlaStatus(Enum):
        OK = "OK"
        WARNING = "Warning"
        ALERT = "Alert"
        ESCALATION = "Escalation"
        NA = "NA"

    class SlaOutcome(Enum):
        UNKNOWN = "unknown"
        MET = "met"
        OVERDUE = "overdue"
        NA = "NA"

    class Prop(Enum):
        SUBJECT = "subject"
        ID = "id"
        PRIORITY = "priority"
        STATUS = "status"
        DOMAIN_NAME = "domain_name"
        SLA_STATUS = "sla_status"
        SLA_OUTCOME = "sla_outcome"
        COMMENTS = "comment"
        REQUESTER = "requester"
        REQUESTER_ID = "requester_id"
        STEPS = "steps"
        CURRENT_STEP = "current_step"
        WORKFLOW = "workflow"
        APPLICATION_DETAILS = "application_details"
        EXPIRATION_DATE = "expiration_date"
        EXPIRATION_FIELD_NAME = "expiration_field_name"

    subject: Optional[str] = prop("No Subject")
    id: Optional[int] = prop(None, cmp=False, jsonify=False)
    workflow: Optional[Workflow] = prop(None, repr=False)
    steps: List[Step] = prop(factory=list, repr=False, flatify=Step.Meta.ROOT.value)
    status: Optional[TicketStatus] = prop(None, jsonify=False)
    domain_name: Optional[str] = prop(None, repr=False)
    sla_status: Optional[SlaStatus] = prop(None, repr=False, jsonify=False)
    sla_outcome: Optional[SlaOutcome] = prop(None, repr=False, jsonify=False)
    expiration_field_name: Optional[str] = prop(None, repr=False, jsonify=False)
    comments: List[Comment] = prop(None, jsonify=False, flatify=Prop.COMMENTS.value)
    _current_step: Optional[Step] = None
    requester: Optional[str] = prop(None, repr=False, jsonify=False)
    requester_id: Optional[int] = prop(None, repr=False)
    expiration_date: Optional[datetime] = prop(
        None,
        repr=False,
        jsonify=False,
        kwargify=lambda val: safe_date(val, TimeFormat.DATE),
    )
    application_details: Optional[ApplicationDetails] = prop(
        None, repr=False, jsonify=False
    )
    priority: Optional[Priority] = prop(Priority.NORMAL, repr=False)
    data: Optional[dict] = attr.ib(None, repr=False)
    _json_override: dict = attr.ib(None, repr=False, eq=False, init=False)

    @classmethod
    def create(cls, workflow: str, subject: Optional[str] = None):
        ticket = cls(subject=subject)
        ticket.workflow = Workflow(name=workflow)

        return ticket

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)
        inst = cls(**kwargs)
        inst.current_step = _obj.get("current_step")
        return inst

    @property
    def current_step_name(self) -> Optional[str]:
        if self.current_step:
            return self.current_step.name
        return None

    @property
    def _json(self) -> dict:
        return (
            self._json_override
            if self._json_override is not None
            else {self.Meta.ROOT.value: jsonify(self)}
        )

    @_json.setter
    def _json(self, val) -> None:
        self._json_override = val

    @property
    def current_step_index(self) -> Optional[int]:
        if not self._current_step:
            return None

        for idx, s in enumerate(self.steps):
            if s.id == self._current_step.id:
                return idx

    @property
    def previous_step(self) -> Optional[Step]:
        current_step_idx = self.current_step_index
        if current_step_idx is None or current_step_idx == 0:
            return None
        else:
            return self.steps[current_step_idx - 1]

    @property
    def current_step(self) -> Optional[Step]:
        return self._current_step

    @current_step.setter
    def current_step(self, step):
        if isinstance(step, dict):
            # extract the current_step name from the API output
            step = step.get("name")
        current_step = self.get_step(step)
        if current_step:
            self._current_step = current_step

    def create_step(self, name) -> Step:
        step = Step(name=name)
        self.steps.append(step)
        return step

    def get_step(self, identifier: Union[int, str]) -> Optional[Step]:
        step = None
        if isinstance(identifier, int):
            try:
                step = self.steps[identifier]
            except IndexError:
                return None
        else:
            for step in self.steps:
                if step.name == identifier:
                    break
            else:
                return None
        return step

    def _get_step(self, step_id: int) -> Optional[Step]:
        step = [s for s in self.steps if s.id == step_id]
        return step[0] if step else None

    def get_step_task(
        self, identifier: Union[int, str], task: Union[int, str] = 0
    ) -> Optional[Task]:
        step = self.get_step(identifier)
        return step.get_task(task) if step else None

    def get_step_task_fields(
        self,
        identifier: Union[int, str],
        task: Union[int, str] = -1,
        *types: Type[Field],
    ) -> Optional[list]:
        _task = self.get_step_task(identifier, task)
        if _task:
            return _task._get_fields(*types)
        return None

    @property
    def last_step(self) -> Step:
        return self.steps[-1]

    @property
    def last_task(self) -> Task:
        return self.last_step.last_task

    @property
    def current_task(self) -> Optional[Task]:
        if self.current_step:
            if len(self.current_step.tasks) > 1:
                raise ValueError(
                    "Ticket.current_task cannot be used when the current step has more than one task, use Ticket.current_tasks instead"
                )
        return self.current_step.last_task if self.current_step else None

    @property
    def current_tasks(self) -> List[Task]:
        return self.current_step.tasks if self.current_step else []

    @property
    def post_json(self) -> dict:
        _self = deepcopy(self)
        _self.steps = [self.steps[0]]
        return _self._json

    @property
    def group_modify(self) -> Optional[MultiGroupChange]:
        if self.current_task:
            return self.current_task.group_modify
        return None

    @property
    def access_request(self) -> Optional[MultiAccessRequest]:
        if self.current_task:
            return self.current_task.access_request
        return None

    @property
    def ar(self) -> Optional[MultiAccessRequest]:
        return self.access_request

    def add_comment(
        self,
        comment: str,
        attachment_uuids: List[str] = None,
        step_id: Union[int, str] = None,
        task_id: Union[int, str] = None,
    ) -> str:
        if step_id is None:
            step_id = self.current_step.id

        if task_id is None:
            task_id = self.current_task.id
        return Scw.add_comment(
            Scw.default, self.id, step_id, task_id, comment, attachment_uuids
        )

    def add_comments(
        self,
        comments: List[Comment],
        step_id: Union[int, str] = None,
        task_id: Union[int, str] = None,
    ) -> List[str]:
        return [
            self.add_comment(
                comment.content,
                [a.uid for a in comment.attachments] if comment.attachments else None,
                step_id,
                task_id,
            )
            for comment in comments
        ]

    def delete_comment(
        self,
        comment_id: Union[int, str] = None,
    ):
        Scw.delete_comment(Scw.default, self.id, comment_id)

    def delete_comments(
        self,
        comment_ids: List[Union[int, str]],
    ):
        for id in comment_ids:
            self.delete_comment(id)

    def get_history(self):
        return Scw.get_ticket_history(Scw.default, self.id)

    def post(self) -> "Ticket":
        res = Scw.default.api.session.post("tickets", json=self.post_json)
        if res.ok:
            created_url = res.headers.get("Location", "")
            tid: str = created_url.split("/")[-1]
            return Scw.default.get_ticket(int(tid))
        else:
            res.raise_for_status()

    def advance(
        self, save: bool = True, silence: Optional[bool] = False
    ) -> Union["Ticket", bool]:
        if self.current_step:
            self.current_step.done()
            if not silence:
                return self.save()  # type: ignore
            else:
                self.save(silence=True)
        else:
            raise AssertionError(
                f"Ticket is currently {self.status}, and cannot be modified"
            )

    def reject(
        self,
        handler_id: Optional[int] = None,
        comment: Optional[str] = "",
    ) -> "Ticket":
        if not self.id:
            raise ValueError("You cannot reject a ticket with no id")
        r = Scw.default.api.reject_ticket(
            self.id, comment=comment, handler_id=handler_id
        )
        if not r.ok:
            try:
                response_body = r.json()
                r.raise_for_status()
            except HTTPError as e:
                msg = response_body.get("result").get("message")
                raise HTTPError(f"Got msg: {msg} from error :{e}")
            except ValueError as e:
                raise ValueError(e)

        else:
            return Scw.default.get_ticket(self.id)

    def cancel(
        self,
        requestor_id: Optional[int] = None,
    ) -> "Ticket":
        r = Scw.default.api.cancel_ticket(self.id, requestor_id=requestor_id)
        if not r.ok:
            msg = safe_unwrap_msg(r)
            raise ValueError(f"Got error: {msg} with status code {r.status_code}")
        return Scw.default.get_ticket(self.id)

    def confirm(
        self,
        requestor_id: Optional[int] = None,
        comment: Optional[str] = "",
    ) -> "Ticket":
        r = Scw.default.api.confirm_ticket(
            self.id, comment=comment, requestor_id=requestor_id
        )
        if not r.ok:
            msg = safe_unwrap_msg(r)
            raise ValueError(f"Got error: {msg} with status code {r.status_code}")
        return Scw.default.get_ticket(self.id)

    def save(
        self, force: Optional[bool] = False, silence: Optional[bool] = False
    ) -> "Ticket":
        if not self.current_step:
            raise AssertionError(
                f"Ticket status is {self.status}, and cannot be modified"
            )
        if not self.id:
            raise AssertionError("Cannot save new Ticket, use Ticket.post() instead")
        for task in self.current_step.tasks:
            res = None
            if not task.id:
                continue
            if force:
                res = Scw.default.api.put_task(self.id, task.id, task._json)
            elif task._dirty:
                res = Scw.default.api.put_task(self.id, task.id, save_task_body(task))
            elif task._dirty_fields:
                res = Scw.default.api.put_fields(
                    self.id, task.id, save_fields_body(task)
                )
            if res is not None and not res.ok:
                try:
                    res.raise_for_status()
                except HTTPError as e:
                    msg = safe_unwrap_msg(res)
                    raise ValueError(
                        f"Status: '{res.status_code}' Message: '{msg}' from API.  Exception: {e}"
                    )
        if not silence:
            return Scw.default.get_ticket(self.id)

    def change_requester(
        self,
        user: Union[SCWUser, int, str],
        comment: str = "",
    ) -> None:
        return Scw.default.change_requester(self, user=user, comment=comment)

    def reassign(
        self,
        user: Union[SCWUser, int, str],
        step: Union[None, Step, int, str] = None,
        task: Union[None, Task, int] = None,
        comment: str = "",
    ) -> None:
        return Scw.default.reassign_ticket(
            self, user=user, step=step, task=task, comment=comment
        )

    def redo(self, step: Union[Step, int, str], comment: str = "") -> "Ticket":
        if len(self.steps) < 2:
            raise AssertionError("Cannot redo a ticket on the first step")
        if not self.current_step:
            raise AssertionError(f"Cannot redo ticket with status {self.status}")
        if not isinstance(step, Step):
            _step = self.get_step(step)
            if not _step:
                raise IndexError(f"Step {step} not found")
            step = _step
        if not step.id:
            raise AssertionError(f"Step {step.name} has no id")
        if step == self.current_step:
            raise AssertionError(f"Cannot redo to current step {step.name}")
        if not self.id:
            raise AssertionError("Cannot redo ticket with no id")
        if not self.current_step.id:
            raise AssertionError("Cannot redo ticket at step with no id")
        if not self.current_step.last_task.id:
            raise AssertionError("Cannot redo ticket at task with no id")
        r = Scw.default.api.redo_step(
            self.id,
            self.current_step.id,
            self.current_step.last_task.id,
            step.id,
            comment,
        )
        if r.ok:
            ticket = Scw.default.get_ticket(self.id)
            count = 0
            while ticket.current_step.id != step.id:  # type: ignore
                count += 1
                if count > 20:  # pragma: no cover
                    raise TimeoutError("Timed out getting ticket after redo")
                sleep(0.5)
                ticket = Scw.default.get_ticket(self.id)

            return ticket

        else:
            r.raise_for_status()

    def map_rules(
        self,
        handler_id: int = None,
    ) -> None:
        Scw.default.map_rules(self.id, handler_id)

    def designer_redesign(self) -> None:
        Scw.default.designer_redesign(
            self.id,
            self.current_step.id,
            self.current_task.id,
            self.current_task.access_request.id,
        )

    def designer_device_commit(self, device_id) -> None:
        Scw.default.designer_device_commit(
            self.id,
            self.current_step.id,
            self.current_task.id,
            self.current_task.access_request.id,
            device_id,
        )

    def designer_device_update(self, device_id) -> None:
        Scw.default.designer_device_update(
            self.id,
            self.current_step.id,
            self.current_task.id,
            self.current_task.access_request.id,
            device_id,
        )


def save_task_body(task: Task) -> dict:
    put_task = deepcopy(task)
    put_task.fields = put_task._dirty_fields

    return {Task.Meta.ROOT.value: put_task._json}


def save_fields_body(task: Task) -> dict:
    put_task = deepcopy(task)
    put_task.fields = put_task._dirty_fields
    _json = put_task._json
    return {Task.Prop.FIELDS.value: _json[Task.Prop.FIELDS.value]}


class TicketIterator(Iterator):
    def __init__(self, session, params):
        self.params = params
        self.data = None
        self.data_index = 0
        self.next_page_link = None
        self.reached_last_page = False
        self.has_error_occurred = False

        self.session = session

    def __iter__(self):
        return self

    def fetch(self):
        if self.next_page_link is not None:
            response = self.session.get(self.next_page_link)
            if response.ok:
                self.has_error_occurred = False
                self.process_response(response)
            else:
                response.raise_for_status()
        elif self.reached_last_page:
            self.data = None
        else:
            response = self.session.get("tickets", params=self.params)

            if response.ok:
                self.has_error_occurred = False
                self.process_response(response)
            else:
                response.raise_for_status()

        self.data_index = 0
        return True

    def process_response(self, response):
        _json = Traverser(response.json())

        self.next_page_link = _json.tickets.get("next", {"@href": None})["@href"]
        data = get_api_node(_json(), "tickets.ticket", listify=True)
        self.data = [Ticket.kwargify(ticket) for ticket in data]
        self.reached_last_page = self.next_page_link is None

    def __next__(self):
        if self.has_error_occurred:
            raise StopIteration

        if self.data is None:
            res = self.fetch()
            if not res:
                self.data = None
                self.has_error_occurred = True
                return res
        elif self.data_index >= len(self.data):
            if self.next_page_link:
                res = self.fetch()

                if not res:
                    self.data = None
                    self.has_error_occurred = True
                    return res
            else:
                raise StopIteration

        if not self.data:
            raise StopIteration

        ret = self.data[self.data_index]
        self.data_index += 1
        return ret


@propify
class CurrentTicketTask(Jsonable):
    id: int = prop(0)
    name: str = prop(None)
    participants: List[dict] = prop(factory=list, flatify="participant")
    status: Optional[dict] = prop(None)


@propify
class CurrentTicketStage(Jsonable):
    id: int = prop(0)
    tasks: List[CurrentTicketTask] = prop(factory=list, flatify="task")


@propify
class TicketSearchResult(Jsonable):
    """
    This class represents a SecureChange ticket search result object returned by tickets/search_by/details
    """

    class Prop(Enum):
        REQUESTER_NAME = "requesterName"
        WORKFLOW_NAME = "workflowName"
        ASSIGNED_TO_NAME = "assignedToName"
        DOMAIN_NAME = "domainName"

    current_step: Optional[str] = prop(None)
    sla_status: Optional[Ticket.SlaStatus] = prop(None)
    requester_name: Optional[str] = prop(None, key=Prop.REQUESTER_NAME.value)
    workflow_name: Optional[str] = prop(None, key=Prop.WORKFLOW_NAME.value)
    assigned_to_name: Optional[str] = prop(None, key=Prop.ASSIGNED_TO_NAME.value)
    domain_name: Optional[str] = prop(None, key=Prop.DOMAIN_NAME.value)
    duration: Optional[int] = prop(None)
    subject: Optional[str] = prop(None)
    status: Optional[TicketStatus] = prop(None)
    priority: Optional[Ticket.Priority] = prop(None)
    id: int = prop(0)

    current_stage: Optional[CurrentTicketStage] = prop(None)

    @property
    def task_status(self) -> Optional[Task.Status]:
        if self.current_stage and self.current_stage.tasks:
            status = self.current_stage.tasks[0].status.get("status", None)
            if status:
                return Task.Status(status)
        return None

    def get_full_ticket(self) -> Ticket:
        return Scw.default.get_ticket(self.id)


# @propify
# class TicketSearchByFreeText(Jsonable):
#     """
#     This class represents a SecureChange ticket search result object returned by tickets/search_by/free_text
#     """

#     class Prop(Enum):
#         workflow_name = "workflowName"
#         domain_name = "domainName"
#         assigned_group = "assignedGroup"


#     assigned_group: Optional[str] = prop(None, key=Prop.assigned_group.value)
#     current_step: Optional[str] = prop(None)
#     sla_status: Optional[Ticket.SlaStatus] = prop(None)
#     requester_name: Optional[str] = prop(None, key=Prop.REQUESTER_NAME.value)
#     workflow_name: Optional[str] = prop(None, key=Prop.workflow_name.value)
#     assigned_to_name: Optional[str] = prop(None, key=Prop.ASSIGNED_TO_NAME.value)
#     domain_name: Optional[str] = prop(None, key=Prop.domain_name.value)
#     duration: Optional[int] = prop(None)
#     subject: Optional[str] = prop(None)
#     status: Optional[TicketStatus] = prop(None)
#     priority: Optional[Ticket.Priority] = prop(None)
#     id: int = prop(0)

#     task_status: Optional[Task.Status] = prop(None)


@propify
class TicketEventParticipant(Jsonable):
    participant_id: int = prop(0)
    participant_name: str = prop(None)


@propify
class TicketEventTaskAssignedData(Jsonable):
    assignee_id: int = prop(0)
    assignee_name: str = prop(None)
    participant_id: int = prop(0)
    participant_name: str = prop(None)


@propify
class TicketEvent(Jsonable):
    """
    This class represents a SecureChange ticket event object returned by tickets/lifecycle_events
    """

    ticket_id: int = prop(0)
    workflow_name: str = prop(None)
    domain_id: int = prop(0)
    parent_workflow_id: int = prop(0)
    timestamp: datetime = prop(None, kwargify=safe_iso8601_date)
    type: str = prop(None)
    step_id: Optional[int] = prop(None)
    step_name: Optional[str] = prop(None)
    task_id: Optional[int] = prop(None)
    task_name: Optional[str] = prop(None)
    potential_participants: List[TicketEventParticipant] = prop(factory=list)
    task_assigned_data: TicketEventTaskAssignedData = prop(None)



from datetime import datetime

from typing import Optional, List, Union
from enum import Enum

from pytos2.utils import propify, prop, TimeFormat
from pytos2.models import Jsonable, UnMapped

from pytos2.securetrack.device import Device
from pytos2.securechange.rule import SlimRule


class VerificationStatus(Enum):
    IMPLEMENTED = "implemented"
    NOT_IMPLEMENTED = "not implemented"


@propify
class Binding(Jsonable):
    class Prop(Enum):
        XSI_TYPE = "@xsi.type"

    class XsiType(Enum):
        POLICY_BINDING = "policy__binding"
        ZONE_BINDING = "zone__binding"
        ZONE_WITH_POLICY_BINDING = "zone__with__policy__binding"
        ACL_BINDING = "acl__binding"
        NON_EXISTING_BINDING = "non_existing_binding"

    xsi_type: Optional[XsiType] = prop(None, key=Prop.XSI_TYPE.value)
    installed_on_module: List[str] = prop(factory=list)


@propify
class NonExistingBinding(Binding):
    incoming_interface_name: Optional[str] = prop(None)
    outgoing_interface_name: Optional[str] = prop(None)


@propify
class PolicyBinding(Binding):
    policy_name: Optional[str] = prop(None)


@propify
class ZoneWithPolicyBinding(PolicyBinding):
    pass


@propify
class ZoneBinding(Binding):
    from_zone: Optional[str] = prop(None)
    to_zone: Optional[str] = prop(None)

    pass


@propify
class ACLBinding(Binding):
    acl_name: Optional[str] = prop(None)
    incoming_interface_name: Optional[str] = prop(None)
    outgoing_interface_name: Optional[str] = prop(None)


def classify_binding(binding):
    binding_types = {
        Binding.XsiType.POLICY_BINDING.value: PolicyBinding,
        Binding.XsiType.ZONE_BINDING.value: ZoneBinding,
        Binding.XsiType.ZONE_WITH_POLICY_BINDING.value: ZoneWithPolicyBinding,
        Binding.XsiType.ACL_BINDING.value: ACLBinding,
        Binding.XsiType.NON_EXISTING_BINDING.value: NonExistingBinding,
    }

    if binding["@xsi.type"] in binding_types:
        cls = binding_types[binding["@xsi.type"]]
        return cls.kwargify(binding)
    else:
        return Binding.kwargify(binding)


@propify
class VerifierBinding(Jsonable):
    class ImplicitCleanupRule(Enum):
        VIOLATED = "VIOLATED"
        NOT_HANDLED = "NOT_HANDLED"
        IMPLEMENTED = "IMPLEMENTED"

    binding: Optional[Binding] = prop(None, kwargify=classify_binding)
    verified: bool = prop(False)
    verifier_error_code: Optional[str] = prop(None)
    verifier_warning: Optional[str] = prop(None)
    percent_implemented: int = prop(0)
    implementation_percentage_threshold: int = prop(0)
    implementing_rules: List[SlimRule] = prop(factory=list, flatify="implementing_rule")
    violating_rules: List[SlimRule] = prop(factory=list, flatify="violating_rule")
    handled_by_implicit_cleanup_rule: ImplicitCleanupRule = prop("")


@propify
class VerifierTarget(Jsonable):
    management_name: Optional[str] = prop(None)
    management_id: Optional[int] = prop(None)
    device_type: Optional[str] = prop(None)
    revision_number: Optional[int] = prop(None)
    administrator: Optional[str] = prop(None)
    vendor: Optional[Device.Vendor] = prop(None)
    verification_status: Optional[VerificationStatus] = prop(None)

    date: Optional[datetime.date] = prop(
        None, repr=False, kwargify=lambda d: datetime.strptime(d, TimeFormat.DATE.value)
    )
    time: Optional[datetime.time] = prop(
        None, repr=False, kwargify=lambda d: datetime.strptime(d, TimeFormat.TIME.value)
    )

    verifier_error_code: Optional[str] = prop(None)
    verifier_warning: Optional[str] = prop(None)

    verifier_bindings: List[VerifierBinding] = prop(
        factory=list, flatify="verifier_binding"
    )


@propify
class VerifierResultLinkInner(Jsonable):
    class Prop(Enum):
        XSI_TYPE = "@xsi.type"
        HREF = "@href"

    xsi_type: str = prop(key=Prop.XSI_TYPE.value)
    href: str = prop(key=Prop.HREF.value)


@propify
class AccessRequestVerifierResult(Jsonable):
    verifier_targets: List[VerifierTarget] = prop(
        factory=list, flatify="verifier_target"
    )
    topology_image_link: str = prop("", flatify="@href")


@propify
class VerifierResultLink(Jsonable):
    status: Optional[VerificationStatus] = prop(None)
    result: Optional[VerifierResultLinkInner] = prop(None)


def classify_verifier_result(obj):
    if "access_request_verifier_result" in obj:
        return AccessRequestVerifierResult.kwargify(
            obj["access_request_verifier_result"]
        )
    else:
        return UnMapped()



from typing import Optional, List, Union
from enum import Enum

from pytos2.utils import deprecated_property, propify, prop
from pytos2.models import Jsonable

from .rule import SlimRule

from .network_object import (
    classify_object_type,
    NetworkObject,
    HostObject,
    SubnetObject,
    RangeObject,
    GroupObject,
)

from .designer_verifier_common import (
    ServiceObject,
    TransportService,
    IPService,
    ICMPService,
    AnyApplicationService,
    ServiceGroup,
    classify_service_type,
)


@propify
class Zone(Jsonable):
    class Prop(Enum):
        GLOBAL = "global"
        TYPE = "type"

    name: Optional[str] = prop(None)
    _type: Optional[str] = prop(
        None, key=Prop.TYPE.value, cmp=False, repr=False, init=False
    )
    _global: Optional[bool] = prop(
        None, key=Prop.GLOBAL.value, cmp=False, repr=False, init=False
    )


@propify
class Instruction(Jsonable):
    class InstructionType(Enum):
        NA = "NA"
        NEW_RULE = "NEW_RULE"
        ADD_OBJECT_TO_DEVICE = "ADD_OBJECT_TO_DEVICE"
        UPDATE_RULE = "UPDATE_RULE"
        UPDATE_GROUP = "UPDATE_GROUP"
        REMOVE_RULE = "REMOVE_RULE"
        RULE_REPLACE = "RULE_REPLACE"
        FULLY_IMPLEMENTED = "FULLY_IMPLEMENTED"
        NO_SECURITY_ON_INTERFACE = "NO_SECURITY_ON_INTERFACE"
        NO_NEED_TO_CHANGE = "NO_NEED_TO_CHANGE"

    class RulePlacement(Enum):
        BEFORE = "BEFORE"
        AFTER = "AFTER"

    class ChangeAction(Enum):
        ADD = "ADD"
        REMOVE = "REMOVE"

    change_action: Optional[ChangeAction] = prop(None)
    rule: Optional[SlimRule] = prop(None)

    rule_order: Optional[str] = prop(None)
    rule_placement: Optional[RulePlacement] = prop(None)

    implements_access_requests: List[str] = prop(
        factory=list, flatify="order", repr=True
    )
    status: Optional[str] = prop(None)
    comment: Optional[str] = prop(None)
    instruction_type: Optional[InstructionType] = prop(None)

    sources: List[NetworkObject] = prop(
        factory=list,
        repr=False,
        kwargify=classify_object_type,
        flatify="source_object",
        key="source_objects",
    )
    destinations: List[NetworkObject] = prop(
        factory=list,
        repr=False,
        kwargify=classify_object_type,
        flatify="destination_object",
        key="destination_objects",
    )

    services: List[ServiceObject] = prop(
        factory=list,
        repr=False,
        kwargify=classify_service_type,
        flatify="service_object",
        key="service_objects",
    )

    source_objects = deprecated_property("source_objects", "sources")
    destination_objects = deprecated_property("destination_objects", "destinations")


Rule = SlimRule


@propify
class ErrorInstruction(Instruction):
    error_message: Optional[str] = prop(None)


@propify
class NoSecurityInstruction(Instruction):
    pass


@propify
class FullyImplementedInstruction(Instruction):
    pass


@propify
class AddServiceObjectInstruction(Instruction):
    class Prop(Enum):
        DeviceAddedServiceObject = "device_added_service_object"

    object: Union[
        TransportService, IPService, ICMPService, ServiceGroup, AnyApplicationService
    ] = prop(
        None,
        repr=False,
        key=Prop.DeviceAddedServiceObject.value,
        kwargify=classify_service_type,
    )


@propify
class AddNetworkObjectInstruction(Instruction):
    class Prop(Enum):
        DeviceAddedNetworkObject = "device_added_network_object"

    object: Union[HostObject, SubnetObject, RangeObject, GroupObject] = prop(
        None,
        repr=False,
        key=Prop.DeviceAddedNetworkObject.value,
        kwargify=classify_object_type,
    )


@propify
class AddNewRuleInstruction(Instruction):
    pass


@propify
class UpdateGroupInstruction(Instruction):
    modified_object_name: Optional[str] = prop(None)
    modified_object_uid: Optional[str] = prop(None)


@propify
class RemoveRuleInstruction(Instruction):
    modified_object_name: Optional[str] = prop(None)
    modified_object_uid: Optional[str] = prop(None)

    rule_uid: Optional[str] = prop(None)

    original_shadowed_rule_num: Optional[int] = prop(None)


@propify
class UpdateRuleInstruction(Instruction):
    class RulePlacement(Enum):
        ModifiedObjectUID = "modified_object_uid"

    class ChangeAction(Enum):
        ADD = "ADD"

    rule: Optional[SlimRule] = prop(None)

    modified_object_name: Optional[str] = prop(None)
    modified_object_uid: str = prop(None)


def get_instruction_type(instruction: dict):
    if (
        instruction["instruction_type"]
        == Instruction.InstructionType.ADD_OBJECT_TO_DEVICE.value
    ):
        if "device_added_service_object" in instruction:
            return AddServiceObjectInstruction.kwargify(instruction)
        else:
            return AddNetworkObjectInstruction.kwargify(instruction)
    elif instruction["instruction_type"] == Instruction.InstructionType.NEW_RULE.value:
        return AddNewRuleInstruction.kwargify(instruction)
    elif (
        instruction["instruction_type"] == Instruction.InstructionType.REMOVE_RULE.value
    ):
        return RemoveRuleInstruction.kwargify(instruction)
    elif (
        instruction["instruction_type"] == Instruction.InstructionType.UPDATE_RULE.value
    ):
        return UpdateRuleInstruction.kwargify(instruction)
    elif (
        instruction["instruction_type"]
        == Instruction.InstructionType.UPDATE_GROUP.value
    ):
        return UpdateGroupInstruction.kwargify(instruction)
    elif (
        instruction["instruction_type"]
        == Instruction.InstructionType.FULLY_IMPLEMENTED.value
        or instruction["status"] == "DESIGN_FULLY_IMPLEMENTED"
    ):
        return FullyImplementedInstruction.kwargify(instruction)
    elif (
        instruction["instruction_type"]
        == Instruction.InstructionType.NO_SECURITY_ON_INTERFACE.value
    ):
        return NoSecurityInstruction.kwargify(instruction)
    elif instruction["instruction_type"] == Instruction.InstructionType.NA.value:
        return ErrorInstruction.kwargify(instruction)
    else:
        return Instruction.kwargify(instruction)


@propify
class BindingSuggestion(Jsonable):
    binding_uid: Optional[str] = prop(None)
    binding_name: Optional[str] = prop(None)
    instructions: List[Instruction] = prop(
        factory=list, flatify="instruction", repr=True, kwargify=get_instruction_type
    )


@propify
class DeviceSuggestion(Jsonable):
    class Prop(Enum):
        SUGGESTIONS_PER_BINDING = "suggestions_per_binding"

    management_name: Optional[str] = prop(None)
    management_id: Optional[int] = prop(None)
    vendor_name: Optional[str] = prop(None, repr=False)
    revision_number: Optional[int] = prop(None, repr=False)
    offline_device: Optional[bool] = prop(False, repr=False)
    device_software_version: Optional[str] = prop(None, repr=False)

    ancestor_management_id: Optional[int] = prop(None, repr=False)
    ancestor_management_name: Optional[str] = prop(None, repr=False)
    ancestor_management_revision_id: Optional[int] = prop(None, repr=False)
    push_status: Optional[str] = prop(None)
    binding_suggestions: List[BindingSuggestion] = prop(
        factory=list,
        key=Prop.SUGGESTIONS_PER_BINDING.value,
        flatify="binding_suggestion",
        repr=False,
    )


@propify
class DesignerResults(Jsonable):
    class Prop(Enum):
        ID = "id"
        DEVICE_SUGGESTION = "device_suggestion"

    class Meta(Enum):
        ROOT = "designer_results"

    device_suggestions: List[DeviceSuggestion] = prop(
        factory=list, flatify=Prop.DEVICE_SUGGESTION.value, key="suggestions_per_device"
    )

    def get_instruction_by_id(self, id):
        for suggestion in self.device_suggestions:
            for binding_suggestion in suggestion.binding_suggestions:
                for instruction in binding_suggestion.instructions:
                    if instruction.id == id:
                        return instruction



from typing import Optional, List

from enum import Enum

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, kwargify, get_api_node


@propify
class MarketplaceApp(Jsonable):
    class Prop(Enum):
        APP_ID = "appId"

    id: str = prop(key="appId")
    name: str = prop("")
    description: str = prop("", repr=False)
    url: str = prop("")
    version: str = prop("")
    installed: bool = prop(False)
    upsell: bool = prop(False, repr=False)

    @property
    def app_id(self):
        return self.id

    @app_id.setter
    def app_id(self, v):
        self.id = v



import pytest
import os
import json
from pytos2.securechange.risk_results import SecurityPolicyViolation
import responses

from pytos2.securetrack.rule import (
    Documentation,
    SecurityRule,
    Track,
    PolicyXsiType,
    RuleLastUsage,
)
from pytos2.securetrack.network_object import ObjectReference

from pytos2.utils import get_api_node


class TestRule:
    @pytest.fixture
    def all_rules(self):
        json_dir = "tests/securetrack/json/rules"

        files = os.listdir(json_dir)
        files = [
            name
            for name in files
            if name.startswith("device-") and name.endswith(".json")
        ]

        json_list = []
        for name in files:
            json_list.append(json.load(open(os.path.join(json_dir, name), "r")))
        # json_list = [json.load(open(os.path.join(json_dir, name), "r")) for name in files]
        return json_list

    def get_rules(self, device_id):
        j = json.load(
            open(f"tests/securetrack/json/rules/device-{device_id}.json", "r")
        )

        return get_api_node(j, "rules.rule", listify=True)

    def get_rule(self, device_id, rule_id):
        rules = self.get_rules(device_id)
        for r in rules:
            if int(r["id"]) == int(rule_id):
                return r

        return None

    @pytest.fixture
    def rules_20(self):
        return self.get_rules(20)

    @pytest.fixture
    def rule_174_50(self):
        return self.get_rule(174, 50)

    def test_attributes_2(self, rule_174_50):
        rule = SecurityRule.kwargify(rule_174_50)

        assert rule.name == "Access From AWS"
        assert rule.options[0] == SecurityRule.Option.LOG_FORWARDING
        assert rule.options[1] == SecurityRule.Option.LOG_SESSION

        assert isinstance(rule.applications[0], ObjectReference)
        assert rule.applications[0].uid == "{bf62fdcf-e2c8-442e-a455-8bbc070e1cc3}"
        assert rule.applications[0].display_name == "Any"

        assert isinstance(rule.users[0], ObjectReference)
        assert rule.users[0].uid == "{bf62fdcf-e2c8-442e-a455-8bbc070e1cc3}"
        assert rule.users[0].display_name == "Any"

        assert isinstance(rule.additional_parameters[0], ObjectReference)
        assert (
            rule.additional_parameters[0].uid
            == "{8abbaf52-4dad-f129-c503-4dd83927276d}"
        )
        assert rule.additional_parameters[0].display_name == "Cloud"
        assert rule.additional_parameters[0].name == "Cloud(tag)"

        assert rule.rule_type == SecurityRule.RuleType.UNIVERSAL

        # Another test for
        # time, application, src_service, dst_service, vpn, install,
        # users, user_access, url_category, src_zone, dst_zone, additional_parameter,

    def test_attributes(self, rules_20):
        rule = rules_20[0]
        rule = SecurityRule.kwargify(rule)

        assert rule.id == 1252
        assert rule.cp_uid == "{EA9DB13E-D058-45C6-A2F0-CD731027C22B}"
        assert rule.uid == "{EA9DB13E-D058-45C6-A2F0-CD731027C22B}"
        assert rule.order == 2
        assert rule.name == ""
        assert rule.comment == "Do not touch ! access to CP !"
        assert rule.action == SecurityRule.Action.ACCEPT
        assert not rule.is_implicit
        assert not rule.is_disabled
        assert not rule.is_external
        assert rule.global_location == SecurityRule.GlobalLocation.MIDDLE
        assert not rule.acceleration_breaker
        assert not rule.is_authentication_rule
        assert rule.rule_number == 1
        assert type(rule.track) is Track
        assert rule.track.level is Track.Level.LOG
        assert len(rule.options) == 0
        assert not rule.src_networks_negated
        assert not rule.dest_networks_negated
        assert not rule.src_services_negated
        assert not rule.dest_services_negated
        assert type(rule.src_networks) is list
        assert type(rule.dest_networks) is list
        assert rule.src_networks[0].id == 3865
        assert rule.src_networks[0].display_name == "Any"
        assert rule.dest_networks[0].id == 3886
        assert rule.dest_networks[0].display_name == "SMC_10.100.200.110"

        assert rule.type == SecurityRule.Type.RULE

        # Bindings checking
        policy = rule.bindings[0].policy
        assert policy.xsi_type == PolicyXsiType.MANAGEMENT_POLICY
        assert policy.id == 17
        assert policy.name == "Standard"
        assert policy.itg_id == 20
        assert policy.itg == "ALL"
        assert policy.is_unique_active_in_itg

        binding = rule.bindings[0]
        assert binding.rule_count == 40
        assert binding.security_rule_count == 37
        assert binding.default
        assert binding.uid == "{3A1BA062-6B19-4C97-8F18-79CBA9EF0AA6}"

        # Additional parameters?

    def test_all_kwargify(self, all_rules):
        for i, _rules in enumerate(all_rules):
            if _rules is None:
                continue

            rules = get_api_node(_rules, "rules.rule", listify=True)
            for rule in rules:
                SecurityRule.kwargify(rule)


class TestRuleLastUsage:
    @responses.activate
    def test_rule_last_usage(self, rule_last_usage_mock, st):
        rule_last_usage_list = st.get_device_rule_last_usage(1)

        assert (
            rule_last_usage_list[0].rule_uid == "89e6b631-705d-4631-a92a-7266e8c31b2f"
        )
        assert not rule_last_usage_list[0].applications
        assert not rule_last_usage_list[0].users
        assert (
            rule_last_usage_list[0].rule_last_hit.strftime("%Y-%m-%d") == "2015-09-21"
        )

        assert (
            rule_last_usage_list[1].rule_uid == "d5adc685-6498-47d0-ad62-f7eeae18a069"
        )
        assert rule_last_usage_list[1].applications == {"facebook-chat": "2018-03-12"}
        assert rule_last_usage_list[1].users == {"tcselab\\ateam3": "2018-03-12"}
        assert (
            rule_last_usage_list[1].rule_last_hit.strftime("%Y-%m-%d") == "2018-03-12"
        )

        rule_last_usage = st.get_device_rule_last_usage_for_uid(
            1, "d5adc685-6498-47d0-ad62-f7eeae18a069"
        )

        assert rule_last_usage.rule_uid == "d5adc685-6498-47d0-ad62-f7eeae18a069"
        assert rule_last_usage.applications == {"facebook-chat": "2018-03-12"}
        assert rule_last_usage.users == {"lab\\ateam3": "2018-03-12"}
        assert rule_last_usage.rule_last_hit.strftime("%Y-%m-%d") == "2018-03-12"


class TestRuleDocumentation:
    @responses.activate
    def test_documentation(self, rule_documentation_mock, st):
        doc = st.get_revision_rule_documentation(24522, 53419)
        assert isinstance(doc, Documentation)

        response = st.update_rule_documentation_by_revision(24522, 53419, doc)
        assert response is None

        response = st.delete_rule_documentation_by_revision(24522, 53419)
        assert response is None

        response = st.delete_rule_documentation(264, 53419)
        assert response is None


class TestUSPViolatingRules:
    @responses.activate
    def test_violating_rules(self, violating_rules_mock, st):
        count = st.get_usp_violating_rules_count(264)
        assert count == 17

        violating_rules = st.get_usp_violating_rules(
            1, SecurityPolicyViolation.Severity.LOW
        )
        assert isinstance(violating_rules, list)
        assert violating_rules[0].id == 787
        assert violating_rules[0].order == 14



from typing import List
import pytest
import json

from pytos2.api import Pager
from pytos2.securetrack.rule import SecurityRule
from . import conftest

from pytos2.securetrack.service_object import (
    Service,
    ServiceXsiType,
    ServiceObjectReference,
    TCPServiceObject,
    UDPServiceObject,
    ICMPServiceObject,
    IPServiceObject,
    PortServiceObject,
    ICMPV6ServiceObject,
    ICMPV6IPServiceObject,
    OtherServiceObject,
    OtherIPServiceObject,
    AnyObject,
    AnyIPServiceObject,
    ServiceGroup,
    RPCServiceObject,
    DCERPCService,
    DefaultService,
    classify_service_object,
)
import responses
from pytos2.utils import get_api_node


class TestAuroraService:
    @pytest.fixture
    def services_list(self):
        j = json.load(open("tests/securetrack/json/services/aurora339.json"))
        service_node = get_api_node(j, "services.service")
        return [classify_service_object(service) for service in service_node]

    def test_service_objects(self, services_list):
        assert len(services_list) == 325

    @responses.activate
    def test_various_services_endpoints(self, services_mock, st):
        pager = st.search_services(uid="3fbd8116-3801-4bee-8593-3cbf999da671")
        assert isinstance(pager, Pager)
        services = pager.fetch_all()
        assert len(services) == 13
        assert isinstance(services[0], Service)

        groups = st.get_service_groups(2961503)
        assert isinstance(groups, list)
        assert len(groups) == 2
        assert isinstance(groups[0], ServiceGroup)

        pager = st.get_service_rules(2937342)
        assert isinstance(pager, Pager)
        rules = pager.fetch_all()
        assert len(rules) == 6
        assert isinstance(rules[0], SecurityRule)

        pager = st.get_services(8, name="daytime")
        assert isinstance(pager, Pager)
        services = pager.fetch_all()
        assert len(services) == 2
        assert isinstance(services[0], Service)

        pager = st.get_services(revision=24522, name="MSExchange")
        assert isinstance(pager, Pager)
        services = pager.fetch_all()
        assert len(services) == 1
        assert isinstance(services[0], Service)

        services = st.get_services(revision=24522, object_ids=[3330108])
        assert len(services) == 4
        assert isinstance(services[0], Service)

        services = st.get_services(8, object_ids=[3099979])
        assert len(services) == 1
        assert isinstance(services[0], Service)


class TestService:
    @pytest.fixture
    def services_list(self):
        j = json.load(open("tests/securetrack/json/services/1.json"))
        service_node = get_api_node(j, "services.service")
        return [classify_service_object(service) for service in service_node]

    def test_attributes(self, services_list):
        assert services_list[0].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[0].id == 2878000
        assert services_list[0].name == "902 (tcp)"
        assert services_list[0].display_name == "902"
        assert services_list[0].class_name == Service.ClassName.TCP_SERVICE
        assert services_list[0].type == Service.Type.TCP_SERVICE
        assert services_list[0].is_global is False
        assert services_list[0].comment == ""
        assert services_list[0].uid == "{4647cde8-ba3f-4721-b40b-818545261f1a}"
        assert services_list[0].overrides is False
        assert services_list[0].min_port == 902
        assert services_list[0].max_port == 902
        assert services_list[0].protocol == Service.Protocol.TCP.value
        assert services_list[0].negate is False
        assert services_list[0].timeout == "0"
        assert services_list[0].is_implicit is True

        assert services_list[2].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[2].id == 2877997
        assert services_list[2].name == "Any"
        assert services_list[2].display_name == "Any"
        assert services_list[2].class_name == Service.ClassName.ANY_OBJECT
        assert services_list[2].type == Service.Type.IP_SERVICE
        assert services_list[2].is_global is False
        assert services_list[2].comment is None
        assert services_list[2].uid == "{97AEB369-9AEA-11D5-BD16-0090272CCB30}"
        assert services_list[2].overrides is False
        assert services_list[2].min_port == 0
        assert services_list[2].max_port == 255
        assert services_list[2].negate is False
        assert services_list[2].timeout == "0"
        assert services_list[2].is_implicit is False

        assert services_list[5].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[5].id == 2878002
        assert services_list[5].name == "administratively-prohibited (icmp)"
        assert services_list[5].display_name == "administratively-prohibited"
        assert services_list[5].class_name == Service.ClassName.ICMP_SERVICE
        assert services_list[5].type == Service.Type.ICMP_SERVICE
        assert services_list[5].is_global is False
        assert services_list[5].comment == ""
        assert services_list[5].uid == "{ba7802ac-0d69-4ffe-8b5f-30a36e518e9f}"
        assert services_list[5].overrides is False
        assert services_list[5].protocol == Service.Protocol.ICMP.value
        assert services_list[5].min_port == 3
        assert services_list[5].max_port == 3
        assert services_list[5].negate is False
        assert services_list[5].timeout == "0"
        assert services_list[5].is_implicit is False

        assert services_list[6].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[6].id == 2878003
        assert services_list[6].name == "ah"
        assert services_list[6].display_name == "ah"
        assert services_list[6].class_name == Service.ClassName.OTHER_SERVICE
        assert services_list[6].type == Service.Type.IP_SERVICE
        assert services_list[6].is_global is False
        assert services_list[6].comment == "Protocol: ah (51)"
        assert services_list[6].uid == "{3fbd8116-3801-4bee-8593-3cbf999da671}"
        assert services_list[6].overrides is False
        assert services_list[6].min_port == 51
        assert services_list[6].max_port == 51
        assert services_list[6].negate is False
        assert services_list[6].timeout == "0"
        assert services_list[6].is_implicit is False

        assert services_list[10].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[10].id == 2878007
        assert services_list[10].name == "aol (udp)"
        assert services_list[10].display_name == "aol"
        assert services_list[10].class_name == Service.ClassName.UDP_SERVICE
        assert services_list[10].type == Service.Type.UDP_SERVICE
        assert services_list[10].is_global is False
        assert services_list[10].comment == "America On-line"
        assert services_list[10].uid == "{7c8de56b-cdd4-4983-91fc-333402a8e93d}"
        assert services_list[10].overrides is False
        assert services_list[10].protocol == Service.Protocol.UDP.value
        assert services_list[10].min_port == 5190
        assert services_list[10].max_port == 5190
        assert services_list[10].negate is False
        assert services_list[10].timeout == "0"
        assert services_list[10].is_implicit is False

        assert services_list[65].xsi_type == ServiceXsiType.SERVICE_GROUP_OBJECT
        assert services_list[65].id == 2878062
        assert services_list[65].name == "group_ServiceGroup_1"
        assert services_list[65].display_name == "ServiceGroup_1"
        assert services_list[65].class_name == Service.ClassName.SERVICE_GROUP
        assert services_list[65].type == Service.Type.GROUP
        assert services_list[65].is_global is False
        assert services_list[65].comment == ""
        assert services_list[65].uid == "{d199bdbd-2309-4f8e-9695-4cfc8aeec90c}"
        assert services_list[65].is_implicit is False
        assert services_list[65].members[0].id == 2878000
        assert (
            services_list[65].members[0].uid == "{4647cde8-ba3f-4721-b40b-818545261f1a}"
        )
        assert services_list[65].members[0].name == "902 (tcp)"
        assert services_list[65].members[0].display_name == "902"

        assert services_list[414].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[414].id == 2973376
        assert services_list[414].uid == "{97AEB3C6-9AEA-11D5-BD16-0090272CCB30}"
        assert services_list[414].name == "nisplus"
        assert services_list[414].display_name == "nisplus"
        assert services_list[414].class_name == Service.ClassName.RPC_SERVICE
        assert services_list[414].type == Service.Type.RPC_SERVICE
        assert services_list[414].is_global is False
        assert (
            services_list[414].comment
            == "NIS+ later version provides additional security and other facilities"
        )
        assert services_list[414].uid == "{97AEB3C6-9AEA-11D5-BD16-0090272CCB30}"
        assert services_list[414].is_implicit is False

        assert services_list[415].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[415].id == 2981341
        assert services_list[415].uid == "{06179529-f281-4596-bb8f-dcbc63cccf9e}"
        assert services_list[415].name == "application-421"
        assert services_list[415].display_name == "s7"
        assert services_list[415].class_name == Service.ClassName.OTHER_SERVICE
        assert services_list[415].type == Service.Type.OTHER_SERVICE
        assert services_list[415].comment == "Protocol: ARP"
        assert services_list[415].is_global is False
        assert services_list[415].timeout == "never"

        assert services_list[416].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[416].id == 2973422
        assert services_list[416].uid == "{3D0D46B6-4DDB-43E0-9FAA-C969DBC3E19F}"
        assert services_list[416].name == "ALL_DCE_RPC"
        assert services_list[416].display_name == "ALL_DCE_RPC"
        assert services_list[416].class_name == Service.ClassName.DCERPC_SERVICE
        assert services_list[416].type == Service.Type.OTHER_SERVICE
        assert (
            services_list[416].comment
            == "Special Service For Allowing All DCE-RPC Services"
        )
        assert services_list[416].is_global is False
        assert services_list[416].timeout == "never"
        assert services_list[416].is_implicit is False

        assert services_list[417].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[417].id == 3000911
        assert services_list[417].uid == "{5a49039f-15b3-4db0-a05f-c83cfc4e4f33}"
        assert services_list[417].name == "/Common/any_any"
        assert services_list[417].display_name == "any_any"
        assert services_list[417].class_name == Service.ClassName.PORT_SERVICE
        assert services_list[417].type == Service.Type.PORT_SERVICE
        assert services_list[417].comment == ""
        assert services_list[417].is_global is False
        assert services_list[417].min_port == 0
        assert services_list[417].max_port == 65535
        assert services_list[417].timeout == "0"

        assert services_list[418].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[418].id == 3014849
        assert services_list[418].uid == "{70ec3e84-daff-40d0-9d9d-63f2aec64aa5}"
        assert services_list[418].name == "ALL_ICMP6"
        assert services_list[418].display_name == "ALL_ICMP6"
        assert services_list[418].class_name == Service.ClassName.ICMPV6_SERVICE
        assert services_list[418].type == Service.Type.IP_SERVICE
        assert services_list[418].comment == ""
        assert services_list[418].is_global is False
        assert services_list[418].min_port == 58
        assert services_list[418].max_port == 58
        assert services_list[418].timeout == "0"
        assert services_list[418].management_domain == "Amsterdam"
        assert services_list[418].management_domain_securetrack_name == "Amsterdam"

        assert services_list[419].xsi_type == ServiceXsiType.SINGLE_SERVICE_OBJECT
        assert services_list[419].id == 3242792
        assert services_list[419].uid == "{B114786F-03CD-AB2E-0B2F-EF9376CBB3E0}"
        assert services_list[419].name == "st_implicit_app_ping6"
        assert services_list[419].display_name == "st_implicit_app_ping6"
        assert services_list[419].class_name == Service.ClassName.ICMP_V6_SERVICE
        assert services_list[419].type == Service.Type.OTHER_SERVICE
        assert services_list[419].comment == ""
        assert services_list[419].is_global is False
        assert services_list[419].timeout == "never"
        assert services_list[419].management_domain == "Shared"
        assert services_list[419].management_domain_securetrack_name == "QA Pano"

    def test_set_attributes(self, services_list):
        services_list[0].max_port = 904

        j = services_list[0]._json
        assert j["max"] == 904



from enum import Enum
from typing import Optional

from requests import HTTPError, JSONDecodeError

from pytos2.api import BaseAPI, get_app_api_session
from pytos2.utils import setup_logger


LOGGER = setup_logger("sa_api")


class SaAPI(BaseAPI):
    class Meta(Enum):
        PATH = "securechangeworkflow/api/secureapp"
        APP = "SCW"
        TOS2_ENV = "SC_SERVER_SERVICE"

    def __init__(
        self, hostname: Optional[str], username: Optional[str], password: Optional[str]
    ):
        self.hostname, self.username, self.password, session = get_app_api_session(
            app=self, hostname=hostname, username=username, password=password
        )

        super().__init__(session)

    def get_application_identities(self):
        return self.session.get("repository/application_identities")

    def get_applications(self):
        return self.session.get("repository/applications")

    def get_application(self, application_id):
        return self.session.get(f"repository/applications/{application_id}")

    def add_application(self, application):
        apps = [application] if not isinstance(application, list) else application

        r = self.session.post(
            "repository/applications", json={"applications": {"application": apps}}
        )
        return r

    def update_application(self, application_id, application):
        return self.session.put(
            f"repository/applications/{application_id}",
            json={"application": application},
        )

    def update_applications(self, applications):
        return self.session.put(
            "repository/applications/",
            json={"applications": {"application": applications}},
        )

    def delete_application(self, application_id):
        return self.session.delete(f"repository/applications/{application_id}")

    def get_application_connections(self, application_id):
        return self.session.get(f"repository/applications/{application_id}/connections")

    def get_application_connection(self, application_id, connection_id):
        r = self.session.get(
            f"repository/applications/{application_id}/connections/{connection_id}",
        )
        return r

    def get_application_history(
        self,
        application_id,
        start_date=None,
        end_date=None,
        count=None,
        start=None,
        type=None,
        user=None,
        show_members=None,
    ):
        return self.session.get(
            f"repository/applications/{application_id}/history",
            params={
                "start_date": start_date,
                "end_date": end_date,
                "count": count,
                "start": start,
                "type": type,
                "user": user,
                "showMembers": show_members,
            },
        )



import pytest
import json
import responses

from pytos2.securetrack.generic_vpn import GenericVpn
from pytos2.utils import get_api_node


class TestGenericVpn:
    device = json.load(open("tests/securetrack/json/generic_vpns/device-3.json"))
    vpns = device["GenericVpns"]

    vpn24 = json.load(open("tests/securetrack/json/generic_vpns/vpn-24.json"))
    vpn = vpn24["GenericVpn"]

    @responses.activate
    def test_add_generic_vpns_200(self, st):
        """Add one or multiple vpns"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=201,
        )

        """Add single vpn"""
        newVpn = st.add_generic_vpn(
            generic=True,
            device=1,
            interface_name="new33",
            vpn_name=None,
            tunnel_source_ip_addr="3.3.3.33",
            tunnel_dest_ip_addr="1.1.1.11",
        )
        assert newVpn is None

    @responses.activate
    def test_add_generic_vpns_400(self, st):
        """Add one or multiple vpns"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=400,
        )

        """Add single vpn"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_vpn(
                generic=True,
                device=1,
                interface_name="new33",
                vpn_name=None,
                tunnel_source_ip_addr="3.3.3.33",
                tunnel_dest_ip_addr="1.1.1.11",
            )
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_add_generic_vpns_404(self, st):
        """Add one or multiple vpns"""
        responses.add(
            responses.POST,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=404,
        )

        """Add single vpn"""
        with pytest.raises(ValueError) as exception:
            st.add_generic_vpn(
                generic=True,
                device=1,
                interface_name="new33",
                vpn_name=None,
                tunnel_source_ip_addr="3.3.3.33",
                tunnel_dest_ip_addr="1.1.1.11",
            )
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_vpn(self, st, generic_vpn_mock):
        """Get vpn by vpn id"""
        mockRes = get_api_node(self.vpn24, "GenericVpn")
        vpn = GenericVpn.kwargify(mockRes)
        vpnByInt = st.get_generic_vpn(24)
        vpnByStr = st.get_generic_vpn("24")
        assert vpnByInt == vpn
        assert vpnByStr == vpn

        with pytest.raises(ValueError) as exception:
            st.get_generic_vpn(404)
        assert "Not Found" in str(exception.value)

        with pytest.raises(ValueError) as exception:
            st.get_generic_vpn("404")
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_get_generic_vpns(self, st, generic_vpn_mock):
        """Get vpns by device id"""
        vpns = [
            GenericVpn.kwargify(d)
            for d in get_api_node(self.device, "GenericVpns", listify=True)
        ]
        vpnsByInt = st.get_generic_vpns(3, generic=True)
        vpnsByInt2 = st.get_generic_vpns(3, generic=False)
        assert vpnsByInt == vpns
        assert vpnsByInt2 == vpns

        with pytest.raises(ValueError) as exception:
            st.get_generic_vpns(404, generic=True)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_update_generic_vpns_200(self, st):
        """Update one or multiple vpns"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=200,
        )

        """Update single vpn"""
        vpn_to_update = GenericVpn.kwargify(self.vpn)
        newVpn = st.update_generic_vpn(vpn_to_update)
        assert newVpn is None

    @responses.activate
    def test_update_generic_vpns_400(self, st):
        """PUT bad request"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=400,
        )

        with pytest.raises(ValueError) as exception:
            vpn_to_update = GenericVpn.kwargify(self.vpn)
            st.update_generic_vpn(vpn_to_update)
        assert "Bad Request" in str(exception.value)

    @responses.activate
    def test_update_generic_vpns_404(self, st):
        """PUT device id not found"""
        responses.add(
            responses.PUT,
            "https://198.18.0.1/securetrack/api/topology/generic/vpn",
            status=404,
        )

        with pytest.raises(ValueError) as exception:
            vpn_to_update = GenericVpn.kwargify(self.vpn)
            st.update_generic_vpn(vpn_to_update)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_vpn(self, st, generic_vpn_mock):
        """Delete vpn by vpn id"""
        vpnByInt = st.delete_generic_vpn(24)
        assert vpnByInt is None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_vpn(404)
        assert "Not Found" in str(exception.value)

    @responses.activate
    def test_delete_generic_vpns(self, st, generic_vpn_mock):
        """Delete vpns by device id"""
        vpnsByInt = st.delete_generic_vpns(1)
        assert vpnsByInt is None

        with pytest.raises(ValueError) as exception:
            st.delete_generic_vpns(404)
        assert "Not Found" in str(exception.value)



import datetime

import pytest
import responses

from pytos2.securetrack.change_window import ChangeWindow


class TestChangeWindows:
    @responses.activate
    def test_get_change_windows(self, change_windows_mock, st):
        change_windows = st.get_change_windows()

        assert change_windows[0].id == "07c230ce-2dec-4109-a0db-33ff45ba1057"
        assert change_windows[0].name == "Panorama - Change Window"
        assert (
            change_windows[0].description
            == "Change window for Panorama devices, will occur every second weekend"
        )
        assert change_windows[0].domain_id == 2
        assert change_windows[0].enabled

        assert change_windows[1].id == "b0dc4034-6d52-4197-8456-eef694b063e3"
        assert change_windows[1].name == "CP R80.10 - Change Window"
        assert change_windows[1].description == "Change window for R80 mgmt. Console"
        assert change_windows[1].domain_id == 9
        assert not change_windows[1].enabled

    @responses.activate
    def test_get_change_windows_fails(self, change_windows_mock_fails, st):
        with pytest.raises(ValueError) as context:
            st.get_change_windows()

        assert "Failed to get resource with status 403" in str(context.value)

        with pytest.raises(ValueError) as context:
            st.get_change_window_tasks("07c230ce-2dec-4109-a0db-33ff45ba1057")

        assert "Error retrieving" in str(context.value)

        with pytest.raises(ValueError) as context:
            st.get_change_window_task("07c230ce-2dec-4109-a0db-33ff45ba1057", 197)

        assert "Error retrieving" in str(context.value)

    @responses.activate
    def test_get_change_windows_fails_json(self, change_windows_mock_fails_json, st):
        with pytest.raises(ValueError) as context:
            st.get_change_windows()

        assert "Failed to decode response" in str(context.value)

        with pytest.raises(ValueError) as context:
            st.get_change_window_tasks("07c230ce-2dec-4109-a0db-33ff45ba1057")

        assert "Error decoding" in str(context.value)

        with pytest.raises(ValueError) as context:
            st.get_change_window_task("07c230ce-2dec-4109-a0db-33ff45ba1057", 197)

        assert "Error decoding" in str(context.value)

    @responses.activate
    def test_get_change_window_tasks(self, change_windows_mock, st):
        change_window_tasks = st.get_change_window_tasks(
            "07c230ce-2dec-4109-a0db-33ff45ba1057"
        )

        assert change_window_tasks[0].start_date == datetime.datetime(
            2024,
            3,
            29,
            0,
            0,
            1,
            tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
        )
        assert change_window_tasks[0].end_date == datetime.datetime(
            2024,
            3,
            29,
            0,
            2,
            1,
            tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
        )
        assert not change_window_tasks[0].errors

        assert change_window_tasks[1].start_date == datetime.datetime(
            2024,
            3,
            15,
            0,
            0,
            0,
            tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
        )
        assert change_window_tasks[1].end_date == datetime.datetime(
            2024,
            3,
            15,
            0,
            2,
            0,
            tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
        )
        assert (
            change_window_tasks[1].errors[0]
            == "Commit not performed as the Device Group monitoring is stopped"
        )

    @responses.activate
    def test_get_change_window_task(self, change_windows_mock, st):
        change_window_task = st.get_change_window_task(
            "07c230ce-2dec-4109-a0db-33ff45ba1057", 197
        )

        assert change_window_task.start_date == datetime.datetime(
            2024,
            3,
            29,
            0,
            0,
            1,
            tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
        )
        assert change_window_task.end_date == datetime.datetime(
            2024,
            3,
            29,
            0,
            2,
            1,
            tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
        )
        assert not change_window_task.errors

        assert not change_window_task.device_commits[0].result.errors
        assert change_window_task.device_commits[0].result.device.status == "SUCCESS"
        assert change_window_task.device_commits[0].result.device.revision_id == 264
        assert not change_window_task.device_commits[0].result.device.warnings

        assert not change_window_task.device_commits[1].result.errors
        assert (
            change_window_task.device_commits[1].result.device.status
            == "SUCCESS_WITH_WARNINGS"
        )
        assert change_window_task.device_commits[1].result.device.revision_id == 370
        assert (
            change_window_task.device_commits[1].result.device.warnings[0]
            == "Commit not performed as the Device Group monitoring is stopped"
        )

    def test_change_window_uuid_property(self):
        import uuid

        change_window = ChangeWindow(domain_id=8)

        _id = str(uuid.uuid4())
        change_window.id = _id

        assert change_window.id == _id
        assert change_window.uuid == _id

        _id = str(uuid.uuid4())
        change_window.uuid = _id

        assert change_window.id == _id



from typing import Optional, List
from enum import Enum

from pytos2.models import Jsonable
from pytos2.utils import propify, prop
from .domain import Domain

from netaddr import IPAddress, IPNetwork  # type: ignore


@propify
class Zone(Jsonable):
    class Prop(Enum):
        USER_NETWORKS = "userNetworks"

    domain: Optional[Domain] = prop(False, repr=False)
    internet: Optional[bool] = prop(None, repr=False)
    unassociated_networks: Optional[bool] = prop(None, repr=False)
    user_networks: Optional[bool] = prop(None, repr=False, key=Prop.USER_NETWORKS.value)

    importing_domain: Optional[Domain] = prop(None, repr=False)
    comment: str = prop("")
    shared: bool = prop(False, repr=False)
    name: str = prop("")


@propify
class ZoneEntry(Jsonable):
    class Props(Enum):
        ZONE_ID = "zoneId"
        ZONE_NAME = "zoneName"

    id: int = prop(0, repr=False)
    ip: Optional[IPAddress] = prop(None)
    netmask: Optional[IPAddress] = prop(None, repr=False)
    prefix: int = prop(0)
    zone_id: int = prop(0, key=Props.ZONE_ID.value, repr=False)
    zone_name: str = prop("", key=Props.ZONE_NAME.value, repr=False)

    @property
    def subnet(self):
        return IPNetwork(str(self.ip) + "/" + str(self.prefix))


@propify
class ZoneReference(Jsonable):
    """
    For mapping zones/{id}/descendants and zones/{id}/ancestors.
    """

    id: int = prop(0)
    name: str = prop("")
    zones: List["ZoneReference"] = prop(factory=list, flatify="zone")
    clouds: List["Cloud"] = prop(factory=list, flatify="cloud")


@propify
class Subnet(Jsonable):
    id: int = prop()
    netmask: Optional[str] = prop(None)
    ip: Optional[str] = prop(None)
    uid: Optional[str] = prop(None)


@propify
class CloudEntry(Jsonable):
    class Prop(Enum):
        BEHIND_CLOUD = "behindCloud"

    id: int = prop(0)
    netmask: Optional[str] = prop(None)
    ip: Optional[str] = prop(None)
    uid: Optional[str] = prop(None)

    cloud: Optional["Cloud"] = prop(None)


@propify
class Cloud(Jsonable):
    class Prop(Enum):
        SECURITY_LEVEL = "securityLevel"
        CUSTOMER_ID_TAG = "customerIdTag"
        INTERFACE_IP = "interfaceIp"
        CUSTOMER_NAME = "customerName"

    id: int = prop()
    security_level: Optional[dict] = prop(None, key=Prop.SECURITY_LEVEL.value)
    customer_id_tag: Optional[int] = prop(None, key=Prop.CUSTOMER_ID_TAG.value)
    interface_ip: Optional[str] = prop(None, key=Prop.INTERFACE_IP.value)
    subnets: List[Subnet] = prop(factory=list)
    customer_name: Optional[str] = prop(None, key=Prop.CUSTOMER_NAME.value)
    entries: List[CloudEntry] = prop(factory=list)
    uid: Optional[str] = prop(None)



from typing import List, Optional
from enum import Enum

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, deprecated_property

from pytos2.securechange.network_object import (
    classify_object_type,
    NetworkObject,
)


from pytos2.securechange.designer_verifier_common import (
    classify_service_type,
    ServiceObject,
)


@propify
class SlimRule(Jsonable):
    class XsiType(Enum):
        SLIM_RULE_WITH_METADATA = "ns3:slimRuleWithMetadataDTO"

    class Prop(Enum):
        RULE_NUMBER = "ruleNumber"
        SOURCE_NETWORKS = "sourceNetworks"
        SOURCE_SERVICES = "sourceServices"
        DESTINATION_NETWORKS = "destinationNetworks"
        DESTINATION_SERVICES = "destinationServices"

    uid: str = prop("")
    is_disabled: bool = prop(False)
    rule_number: int = prop(0, key=Prop.RULE_NUMBER.value)
    sources: List[NetworkObject] = prop(
        factory=list, key=Prop.SOURCE_NETWORKS.value, kwargify=classify_object_type
    )
    destinations: List[NetworkObject] = prop(
        factory=list, key=Prop.DESTINATION_NETWORKS.value, kwargify=classify_object_type
    )
    source_services: List[ServiceObject] = prop(
        factory=list,
        key=Prop.SOURCE_SERVICES.value,
        kwargify=classify_service_type,
    )
    destination_services: List[ServiceObject] = prop(
        factory=list,
        key=Prop.DESTINATION_SERVICES.value,
        kwargify=classify_service_type,
    )
    track: Optional[dict] = prop(None)
    install_ons: List[dict] = prop(factory=list)
    communities: List[dict] = prop(factory=list)
    times: List[dict] = prop(factory=list)

    from_zone: Optional[str] = prop(None, repr=False)
    to_zone: Optional[str] = prop(None, repr=False)

    source_objects = deprecated_property("source_objects", "sources")
    source_networks = deprecated_property("source_networks", "sources")
    destination_objects = deprecated_property("destination_objects", "destinations")
    destination_networks = deprecated_property("destination_networks", "destinations")
    services = deprecated_property("services", "destination_services")


@propify
class SlimRuleMetadata(Jsonable):
    permisssiveness_level: str = prop("")
    violations: List[str] = prop(factory=list)
    last_hit: str = prop("")
    shadowed_status: str = prop("")
    ticket_ids: List[int] = prop(factory=list)
    business_owners: List[str] = prop(factory=list)
    expirations: List[str] = prop(factory=list)
    applications: List[str] = prop(factory=list)
    last_modified: str = prop("")


@propify
class SlimRuleWithMetadata(SlimRule):
    rule_metadata: Optional[SlimRuleMetadata] = prop(None)



from enum import Enum
from typing import Optional, List

from netaddr import IPAddress  # type: ignore
from attr.converters import optional
import attr

from pytos2.models import Jsonable
from pytos2.utils import propify, prop, kwargify

from pytos2 import securetrack


@propify
class DeviceLicenseStatus(Jsonable):
    class Prop(Enum):
        TYPE = "type"

    expiration: Optional[str] = prop(None, repr=False)
    used: Optional[int] = prop(None, repr=False)
    status: str = prop("")
    sku: Optional[str] = prop(None, repr=False)
    type: Optional[str] = prop(
        None, key=Prop.TYPE.value, cmp=False, repr=True, init=True
    )


@propify
class Device(Jsonable):
    class Model(Enum):
        CP_CLM = "cp_clm"
        CP_MDS = "cp_mds"
        CP_CMA = "cp_cma"
        CP_DOMAIN_R80PLUS = "cp_domain_r80plus"
        CP_SMC_R80PLUS = "cp_smc_r80plus"
        CP_MDS_R80PLUS = "cp_mds_r80plus"
        CP_SMRT_CNTR = "cp_smrt_cntr"
        CP_SMART_ONE = "cp_smart_one"
        ROUTER = "router"
        XR_ROUTER = "xr_router"
        MODULE = "module"
        MODULE_CLUSTER = "module_cluster"
        NETSCREEN = "netscreen"
        NETSCREEN_CLUSTER = "netscreen_cluster"
        JUNOS = "junos"
        JUNOSSTATELESS = "junosStateless"
        NSM = "nsm"
        NSM_DEVICE = "nsm_device"
        NSM_NETSCREEN_ISG = "nsm_netscreen_isg"
        CSM = "csm"
        CSM_OLD = "csm_old"
        CSM_DEVICE = "csm_device"
        CSM_ASA = "csm_asa"
        CSM_FWSM = "csm_fwsm"
        CSM_ROUTER = "csm_router"
        CSM_NEXUS = "csm_nexus"
        CSM_SWITCH = "csm_switch"
        FWSM = "fwsm"
        NEXUS = "nexus"
        FMG = "fmg"
        FMG_VDOM = "fmg_vdom"
        FMG_FW = "fmg_fw"
        FMG_VDOM_MANAGER = "fmg_vdom_manager"
        ASA = "asa"
        PIX = "pix"
        BIGIP = "bigip"
        NEW_BIGIP = "new_bigip"
        STONESOFT_SMC = "stonesoft_smc"
        SINGLE_FW = "single_fw"
        MASTER_ENGINE = "master_engine"
        VIRTUAL_FW = "virtual_fw"
        FW_CLUSTER = "fw_cluster"
        L3_SWITCH = "L3_switch"
        SWITCH = "switch"
        FORTIGATE = "fortigate"
        PALOALTOFW = "PaloAltoFW"
        PANORAMA = "Panorama"
        PANORAMA_DEVICE = "Panorama_device"
        PANORAMA_DEVICE_CLUSTER = "Panorama_device_cluster"
        PANORAMA_NG = "Panorama_ng"
        PANORAMA_DEVICE_GROUP = "Panorama_device_group"
        PANORAMA_NG_FW = "Panorama_ng_fw"
        MCAFEEFW = "mcafeeFW"
        PROXYSG = "proxysg"
        IPTABLES = "iptables"
        NSX_MANAGER = "nsx_manager"
        NSX_T_MANAGER = "nsx_t_manager"
        NSX_FW = "nsx_fw"
        NSX_LRTR = "nsx_lrtr"
        NSX_EDGE = "nsx_edge"
        AWS_MANAGER = "aws_manager"
        AWS_VPC = "aws_vpc"
        AWS_GW_LB = "aws_gw_lb"
        OPENSTACK_MANAGER = "openStack_manager"
        OPENSTACK_REGION = "openStack_region"
        AZURE_RM_FW = "azure_rm_fw"
        AZURE_RM_FW_POLICY = "azure_rm_fw_policy"
        AZURE_RM_FW_POLICY_ROOT = "azure_rm_fw_policy_root"
        AZURE_RM_LB = "azure_rm_lb"
        AZURE_RM_MANAGER = "azure_rm_manager"
        AZURE_RM_VNET = "azure_rm_vnet"
        AZURE_VHUB = "azure_vhub"
        AZURE_VWAN = "azure_vwan"
        FORTIMANAGER = "fortimanager"
        FMG_ADOM = "fmg_adom"
        FMG_FIREWALL = "fmg_firewall"
        ACI = "aci"
        ACI_TENANT = "aci_tenant"
        FMC = "fmc"
        FIREPOWER = "firepower"
        FMC_DOMAIN = "fmc_domain"
        ZIA_CLOUD_FIREWALL = "zia_cloud_firewall"
        AWS_TRANSITGATEWAY = "aws_transitgateway"
        MERAKI = "meraki"
        MERAKI_ORGANIZATION = "meraki_organization"
        MERAKI_NETWORK = "meraki_network"
        GOOGLE_CLOUD_PLATFORM = "google_cloud_platform"
        GOOGLE_VPC = "gcp_vpc"

    class VirtualType(Enum):
        DEVICE = "device"
        VDOM_MANAGER = "vdom_manager"
        CONTEXT = "context"
        VT_STANDALONE = "vt_standalone"
        MASTERENGINE = "MasterEngine"
        MDOM = "mdom"
        MANAGEMENT = "management"
        VSX = "vsx"
        VSX_ROUTER = "vsx_router"

    class Vendor(Enum):
        CISCO = "Cisco"
        FORTINET = "Fortinet"
        NETSCREEN = "Netscreen"
        CHECKPOINT = "Checkpoint"
        PALO_ALTO = "PaloAltoNetworks"
        NEW_F5 = "NewF5"
        F5 = "f5"
        MCAFEE = "Mcafee"
        STONESOFT = "Stonesoft"
        BLUECOAT = "bluecoat"
        GENERIC = "Generic"
        LINUX = "linux"
        VMWARE = "VMware"
        AMAZON = "Amazon"
        OPENSTACK = "OpenStack"
        AZURE = "Azure"
        ZSCALER = "Zscaler"
        GOOGLE = "GoogleCloudPlatform"

    id: int = prop(0, converter=int)
    installed_policy: Optional[str] = prop(None)
    latest_revision: Optional[int] = prop(None, converter=optional(int), repr=False)
    parent_id: Optional[int] = prop(None, converter=optional(int))
    virtual_type: Optional[VirtualType] = prop(None)
    context_name: Optional[str] = prop(None)
    _ip: Optional[IPAddress] = attr.ib(None, repr=False)
    os_version: Optional[str] = prop(None, key="OS_Version")
    domain_name: str = prop(None)
    model: Optional[Model] = prop(None)
    vendor: Optional[Vendor] = prop(None)
    offline: bool = prop(False)
    topology: bool = prop(False)
    module_uid: Optional[str] = prop(None)
    module_type: Optional[str] = prop(None)
    domain_id: Optional[int] = prop(None, converter=optional(int), repr=False)
    name: Optional[str] = prop(None)
    status: str = prop(None)
    licenses: Optional[List[DeviceLicenseStatus]] = prop(factory=list)

    @property
    def parent(self) -> Optional["Device"]:
        if not self.parent_id:
            return None
        return securetrack.St.default.get_device(self.parent_id)

    @property
    def parents(self) -> List["Device"]:
        d = self
        parents: list = []
        while d.parent:
            d = d.parent
            parents.insert(0, d)

        return parents

    @property
    def children(self):
        return [
            d for d in securetrack.St.default.get_devices() if d.parent_id == self.id
        ]

    @property
    def ip(self):
        return self._ip

    @ip.setter
    def ip(self, val):
        self._ip = IPAddress(val)

    @classmethod
    def kwargify(cls, obj):
        _obj, kwargs = kwargify(cls, obj)

        ip = _obj.get("ip")

        ip = IPAddress(ip) if ip is not None else None
        kwargs["ip"] = ip

        # handle non license device result
        if not _obj.get("licenses"):
            kwargs["licenses"] = list()
        else:
            kwargs["licenses"] = [
                DeviceLicenseStatus.kwargify(lic) for lic in _obj.get("licenses")
            ]

        return cls(**kwargs)

    def get_revisions(self):
        return securetrack.St.default.get_revisions(device=self.id)

    def get_rules(self):
        return securetrack.St.default.get_rules(device=self.id)

    def get_nat_rules(self):
        return securetrack.St.default.get_nat_rules(device=self.id)

    def get_network_objects(self, add_parent_objects: Optional[bool] = None):
        return securetrack.St.default.get_network_objects(
            device=self.id, add_parent_objects=add_parent_objects
        )

    @property
    def network_objects(self):
        return self.get_network_objects()

    def get_interfaces(self):
        return securetrack.St.default.get_interfaces(device_id=self.id)

    @property
    def interfaces(self):
        return self.get_interfaces()

    def get_topology_interfaces(self):
        return securetrack.St.default.get_topology_interfaces(device_id=self.id)

    @property
    def topology_interfaces(self):
        return self.get_topology_interfaces()

    def get_bindable_objects(self):
        return securetrack.St.default.get_bindable_objects(device_id=self.id)

    @property
    def bindable_objects(self):
        return self.get_bindable_objects()

    def get_time_objects(self):
        return securetrack.St.default.get_time_objects_by_device(self.id)



import sys
import os

import responses
import pytest
from requests.exceptions import HTTPError
from pytos2.securechange.trigger import (
    on_create,
    on_test,
    on_advance,
    on_redo,
    on_resubmit,
    on_reopen,
    on_automation_failure,
    on_reject,
    on_cancel,
    on_close,
    _get_ticket_info_from_stdin,
    get_ticket_id_from_ticket_info,
    workflow_is,
    step_is,
)
import pytos2.securechange.trigger

from pytos2.securechange.ticket import Ticket
from pytos2.securechange.trigger import Trigger, TEST, CREATE, ADVANCE, NONE
from pytos2.securechange import trigger

from .conftest import MockExit


@pytest.fixture
def nonexistent_ticket_mock(monkeypatch, scw, all_fields_ticket_json):
    with monkeypatch.context() as m:
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/9999",
            status=404,
        )

        m.setattr(
            sys.stdin,
            "read",
            lambda: "<ticket_info><id>9999</id><open_request_stage/><current_stage><name>Duplicate</name></current_stage></ticket_info>",
        )
        m.setattr(sys.stdin, "isatty", lambda: False)
        yield m


@pytest.fixture
def ticket_mock(monkeypatch, scw, all_fields_ticket_json):
    with monkeypatch.context() as m:
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288",
            json=all_fields_ticket_json,
        )

        m.setattr(
            sys.stdin,
            "read",
            lambda: "<ticket_info><id>288</id><open_request_stage/><current_stage><name>Duplicate</name></current_stage></ticket_info>",
        )
        m.setattr(sys.stdin, "isatty", lambda: False)
        yield m


@pytest.fixture
def invalid_stdin_ticket_mock(monkeypatch, scw, all_fields_ticket_json):
    with monkeypatch.context() as m:
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288",
            json=all_fields_ticket_json,
        )

        m.setattr(
            sys.stdin,
            "read",
            lambda: "<ticket__info><id>288</id><open_request_stage/><current_stage><name>Duplicate</name></current_stage></ticket__info>",
        )
        m.setattr(sys.stdin, "isatty", lambda: False)
        yield m


@pytest.fixture
def invalid_xml_ticket_mock(monkeypatch, scw, all_fields_ticket_json):
    with monkeypatch.context() as m:
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288",
            json=all_fields_ticket_json,
        )

        m.setattr(
            sys.stdin,
            "read",
            lambda: '{ "ticket_info": { "id": 288, "open_request_stage": null, "current_stage": { "name": Duplicate" } } }',
        )
        m.setattr(sys.stdin, "isatty", lambda: False)
        yield m


@pytest.yield_fixture
def isatty_ticket_mock(monkeypatch, scw, all_fields_ticket_json):
    with monkeypatch.context() as m:
        responses.add(
            responses.GET,
            "https://198.18.0.1/securechangeworkflow/api/securechange/tickets/288",
            json=all_fields_ticket_json,
        )

        m.setattr(sys.stdin, "read", lambda: "")

        m.setattr(sys.stdin, "isatty", lambda: True)
        yield m


class TestTriggerMethods:
    def test_str(self):
        tr = TEST | CREATE

        assert str(tr) == "Test or Create Triggers"

    def test_repr(self):
        tr = TEST | CREATE

        assert repr(tr) == "TEST|CREATE"

    @responses.activate
    def test_get_ticket_id_from_ticket_info(self, ticket_mock):
        ticket_info = _get_ticket_info_from_stdin()()

        ticket_id = get_ticket_id_from_ticket_info(ticket_info)
        assert ticket_id == 288

    @responses.activate
    def test_get_ticket_id_from_invalid_ticket_info(self, invalid_stdin_ticket_mock):
        ticket_info = _get_ticket_info_from_stdin()()

        ticket_id = get_ticket_id_from_ticket_info(ticket_info)
        assert ticket_id is None

    def test_or_contains(self):
        tr = TEST | CREATE

        assert TEST in tr
        assert CREATE in tr

        tr |= TEST
        assert TEST in tr

        tr2 = TEST | CREATE | ADVANCE

        tr3 = tr | tr2
        tr4 = tr2 | tr

        assert tr2 in tr3
        assert tr2 in tr4

        assert CREATE in tr3
        assert CREATE in tr4

        tr5 = NONE | NONE
        assert CREATE not in tr5

        tr6 = tr | tr
        assert tr6 is tr

        tr7 = TEST | NONE

        assert tr7 == TEST


class TestTriggers:
    def setup_method(self, method):
        pytos2.securechange.trigger.get_ticket_info_from_stdin = (
            _get_ticket_info_from_stdin()
        )

    @responses.activate
    def test_nonexistent_ticket(self, monkeypatch, nonexistent_ticket_mock):
        nonexistent_ticket_mock.setenv("SCW_EVENT", "ADVANCE")
        results = {}

        with pytest.raises(HTTPError):

            @on_advance(when=workflow_is("Fake") or step_is("Fake"))
            def f(ticket):
                results["ticket"] = ticket

    def test_isatty(self, isatty_ticket_mock):
        isatty_ticket_mock.setenv("SCW_EVENT", "ADVANCE")

        ticket_info = _get_ticket_info_from_stdin()()
        assert ticket_info is None

    @responses.activate
    def test_invalid_stdin_ticket(self, invalid_stdin_ticket_mock):
        invalid_stdin_ticket_mock.setenv("SCW_EVENT", "TEST")

        ticket_info = _get_ticket_info_from_stdin()()

        assert ticket_info is None

    @responses.activate
    def test_invalid_xml_ticket(self, invalid_xml_ticket_mock):
        invalid_xml_ticket_mock.setenv("SCW_EVENT", "TEST")

        ticket_info = _get_ticket_info_from_stdin()()

        assert ticket_info is None

    @responses.activate
    def test_step_is_matching(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "ADVANCE")
        results = {}

        @on_advance(when=step_is("No Fields"))
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_step_is_not_matching(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "ADVANCE")
        results = {}

        @on_advance(when=step_is("All Fields"))
        def f(ticket):
            results["ticket"] = ticket

        assert results.get("ticket") is None

    @responses.activate
    def test_workflow_is_matching(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "ADVANCE")
        results = {}

        @on_advance(when=workflow_is("AR all fields"))
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_workflow_is_not_matching(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "ADVANCE")
        results = {}

        @on_advance(when=workflow_is("Fake Workflow"))
        def f(ticket):
            results["ticket"] = ticket

        assert results.get("ticket") is None

    @responses.activate
    def test_cached_std_info(self, ticket_mock):
        get_ticket_info = _get_ticket_info_from_stdin()

        ticket_info = get_ticket_info()
        ticket_id = get_ticket_id_from_ticket_info(ticket_info)
        assert ticket_id == 288

        # Hits the cached std_info if-then
        ticket_info = get_ticket_info()
        ticket_id = get_ticket_id_from_ticket_info(ticket_info)
        assert ticket_id == 288

    @responses.activate
    def test_on_test(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "TEST")
        results = {}

        @on_test
        def f(ticket):
            results["ticket"] = ticket

        assert results.get("ticket") is None

    @responses.activate
    def test_on_create(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "CREATE")
        results = {}

        @on_create
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_on_advance(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "ADVANCE")
        results = {}

        @on_advance
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_on_redo(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "REDO")
        results = {}

        @on_redo
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_on_resubmit(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "RESUBMIT")
        results = {}

        @on_resubmit
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_on_reopen(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "REOPEN")
        results = {}

        @on_reopen
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_on_automation_failure(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "AUTOMATION_FAILED")
        results = {}

        @on_automation_failure
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_on_reject(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "REJECT")
        results = {}

        @on_reject
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_on_cancel(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "CANCEL")
        results = {}

        @on_cancel
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)

    @responses.activate
    def test_on_close(self, ticket_mock):
        ticket_mock.setenv("SCW_EVENT", "CLOSE")
        results = {}

        @on_close
        def f(ticket):
            results["ticket"] = ticket

        assert isinstance(results.get("ticket"), Ticket)


class TestTrigger:
    def test_none(self):
        assert str(Trigger.from_scw_event(123)) == "No Trigger"

    def test_contains_none(self):
        t = trigger.ADVANCE
        assert None not in t
